# -*- coding: utf-8 -*-
"""
Created on Fri Dec  7 13:30:46 2018

@author: stasb
"""

import numpy as np
from numpy.random import rand
import pandas as pd

from .models import base_model
from .corrections import base_correction
from .events import base_event, event_detector
from .scaler import scaler

class base_station_keeping:
    '''
    Class base_station_keeping is base class for all station-keeping
    techniques in OrbiPy.
    
    In astrodynamics, the orbital maneuvers made by thruster burns that 
    are needed to keep a spacecraft in a particular assigned orbit are 
    called orbital station-keeping.
    https://en.wikipedia.org/wiki/Orbital_station-keeping
    
    Common scheme
    -------------
    All impulsive station-keeping techniques for unstable orbits near 
    Lagrange points use common scheme:
        (1) Orbit determination (made by first_correction)
        (2) Propagation up to specified time or event (rev)
        (3) Calculation of delta-v (made by correction) and execution of maneuver
        (4) Repeat (2)-(3) for N times
    '''
    def __init__(self, model, first_correction, correction,
                 rev, events,  maxt,
                 verbose=True):
        if not isinstance(model, base_model):
            raise TypeError('model should be an instance of base_model')
        self.model = model
        if not isinstance(first_correction, base_correction):
            raise TypeError('first_correction should be an instance of base_correction')
        self.first_correction = first_correction
        if not isinstance(correction, base_correction):
            raise TypeError('correction should be an instance of base_correction')
        self.correction = correction
        self.rev = rev
        self.is_rev_event = isinstance(rev, base_event)
        self.events = events
        self.maxt = maxt
        self.verbose = verbose
        self.reset()

    def reset(self):
        self.dvout = []
        self.evout = []
        
    def prop(self, t0, s0, N=10, ret_df=True):
        pass
    
    
class simple_station_keeping(base_station_keeping):
    '''    
    Common scheme
    -------------
    All impulsive station-keeping techniques for unstable orbits near 
    Lagrange points use common scheme:
        (1) Orbit determination (made by first_correction)
        (2) Propagation up to specified time or event (rev)
        (3) Calculation of delta-v (made by correction) and execution of maneuver
        (4) Repeat (2)-(3) for N times
        
    Class simple_station_keeping implements common scheme.
    
    See also
    --------
    base_station_keeping
    '''
    
    def __init__(self,
                 model,
                 first_correction, 
                 correction,
                 rev=np.pi,
                 events = [],
                 maxt=10*np.pi,
                 verbose=True):
        super().__init__(model, first_correction, correction, 
                         rev, events, maxt, verbose)
        
    def reset(self):
        super().reset()
        self.arr = []
        
    def prop(self, t0, s0, N=10, ret_df=True):
        self.reset()
                
        s1 = s0.copy()
        t1 = t0        
        
        if self.is_rev_event:
            events = tuple(self.events)+(self.rev,)
            maxt = self.maxt
        else:
            events = tuple(self.events)
            maxt = self.rev
            
        detector = event_detector(self.model, events)
        if self.verbose: print('Simple station-keeping:', end=' ')
        for i in range(N):
            if i == 0:
                # first correction == orbit determination
                dv = self.first_correction.calc_dv(t1, s1)
            else:
                dv = self.correction.calc_dv(t1, s1)
            s1 += dv
            self.dvout.append([t1, *dv])
            cur_rev, ev = detector.prop(s1, t1, t1+maxt, ret_df=False)
            if ev.shape[0] > 0:
                self.evout.append(ev)
            t1 = cur_rev[-1,0]
            s1 = cur_rev[-1,1:].copy()
#            t1 = ev[-1,2]
#            s1 = ev[-1,3:].copy()
            self.arr.append(cur_rev.copy() if i==N-1 else cur_rev[:-1].copy())
            if self.verbose: print(i, end=' ')
        
        if self.verbose: print()
        arr = np.vstack(tuple(self.arr))
        if len(self.dvout) > 0:
            self.dvout = np.array(self.dvout)
        if len(self.evout) > 0:
            self.evout = np.vstack(tuple(self.evout))
        if ret_df:
            df = self.model.to_df(arr)
            return df
        else:
            return arr

class strict_station_keeping(base_station_keeping):
    '''    
    Common scheme
    -------------
    All impulsive station-keeping techniques for unstable orbits near 
    Lagrange points use common scheme:
        (1) Orbit determination (made by first_correction)
        (2) Propagation up to specified time or event (rev)
        (3) Calculation of delta-v (made by correction) and execution of maneuver
        (4) Repeat (2)-(3) for N times
        
    Class strict_station_keeping implements common scheme.
    
    Parameters
    ----------
    
    model, first_correction correction, rev, events, maxt: see base_station_keeping
    
    maxdv : scalar
        Maximum delta-v allowed. Violation raises exception.
        
    border : iterable of base_event inherited objects
        If any of this events will be reached during propagation step (2),
        exception will be raised.
        
    See also
    --------
    base_station_keeping
    '''
    
    def __init__(self,
                 model,
                 first_correction, 
                 correction,
                 rev=np.pi,
                 events = [],
                 maxt=10*np.pi,
                 maxdv=1e-11,
                 border=[],
                 verbose=True):
        super().__init__(model, first_correction, correction, 
                         rev, events, maxt, verbose)
        self.maxdv = maxdv
        self.border = border
        
    def reset(self):
        super().reset()
        self.arr = []
        
    def prop(self, t0, s0, N=10, ret_df=True):
        self.reset()
                
        s1 = s0.copy()
        t1 = t0        
        
        if self.is_rev_event:
            events = tuple(self.events)+tuple(self.border)+(self.rev,)
            maxt = self.maxt
        else:
            events = tuple(self.events)+tuple(self.border)
            maxt = self.rev
        
        brd_num = len(self.border)
        ev_num = len(events)
        # indexes of border events
        brd_idx = tuple(range(ev_num,ev_num+brd_num))

        detector = event_detector(self.model, events)
        if self.verbose: print('Simple station-keeping:', end=' ')
        for i in range(N):
            if i == 0:
                # first correction == orbit determination
                dv = self.first_correction.calc_dv(t1, s1)
            else:
                dv = self.correction.calc_dv(t1, s1)
                dvnorm = np.linalg.norm(dv) > self.maxdv
                if (dvnorm): # check delta-v by threshold
                    raise RuntimeError('Delta-v (%f) > maxdv (%f)'%(dvnorm, self.maxdv))    
            s1 += dv
            self.dvout.append([t1, *dv])
            cur_rev, ev = detector.prop(s1, t1, t1+maxt, ret_df=False)
            if ev.shape[0] > 0:
                self.evout.append(ev)
                if brd_num:
                    if brd_idx in ev[:,0]:
                        raise RuntimeError('Orbit propagated out of borders')
            t1 = cur_rev[-1,0]
            s1 = cur_rev[-1,1:].copy()
#            t1 = ev[-1,2]
#            s1 = ev[-1,3:].copy()
            self.arr.append(cur_rev.copy() if i==N-1 else cur_rev[:-1].copy())
            if self.verbose: print(i, end=' ')
        
        if self.verbose: print()
        arr = np.vstack(tuple(self.arr))
        if len(self.dvout) > 0:
            self.dvout = np.array(self.dvout)
        if len(self.evout) > 0:
            self.evout = np.vstack(tuple(self.evout))
        if ret_df:
            df = self.model.to_df(arr)
            return df
        else:
            return arr
      
class montecarlo_station_keeping(base_station_keeping):
    '''
    Common scheme
    -------------
    All impulsive station-keeping techniques for unstable orbits near 
    Lagrange points use common scheme:
        (1) Orbit determination (made by first_correction)
        (2) Propagation up to specified time or event (rev)
        (3) Calculation of delta-v (made by correction) and execution of maneuver
        (4) Repeat (2)-(3) for N times
        
    Monte-Carlo station-keeping
    ---------------------------
    Corrections are calculated using states with added random noise but
    maneuvers performed using states without noise. Noise is uniform and
    generated with numpy.random.rand generator (MT19937 algorithm).
    Noise is applied for:
        - spacecraft position: random vector in a ball of unc_pos radius;
        - spacecraft velocity: random vector in a ball of unc_vel radius;
        - delta-v: random vector in a ball which radius is unc_dv 
        percents of delta-v magnitude;
    '''
    def __init__(self, 
                 model,
                 first_correction,
                 correction,
                 unc_pos = (5, 'km'),
                 unc_vel = (5, 'cm/s'),
                 unc_dv  = 5, # percent
                 rev=np.pi,
                 events=[],
                 maxt=10*np.pi,              
                 verbose=True):
        super().__init__(model, first_correction, correction, rev, events, maxt, verbose)        
        self.scale = scaler.from_model(self.model)
        self.unc_pos = unc_pos
        self.unc_vel = unc_vel
        self.unc_dv  = unc_dv
        self.unc_pos_nd = self.scale.convert(self.unc_pos[0], self.unc_pos[1], 'nd')
        self.unc_vel_nd = self.scale.convert(self.unc_vel[0], self.unc_vel[1], 'nd/nd')
#        self.randout = []

    def reset(self):
        super().reset()
        self.randout = []

    @staticmethod
    def rand_ball_one():
        d = rand(3) * 2 - 1
        while d[0]**2+d[1]**2+d[2]**2 > 1:
            d = rand(3) * 2 - 1
        return d
    
    @staticmethod
    def rand_ball_v(v, r):
        d = rand(3) * 2 - 1
        while d[0]**2+d[1]**2+d[2]**2 > 1:
            d = rand(3) * 2 - 1
        return v + d * r
    
    @staticmethod
    def rand_ball_percent(v, r):
        d = rand(3) * 2 - 1
        while d[0]**2+d[1]**2+d[2]**2 > 1:
            d = rand(3) * 2 - 1
        nrm = (v[0]**2+v[1]**2+v[2]**2)**0.5
        return v + d * (nrm * r)

    def prop(self, t0, s0, N=10, ret_df=True):
        self.reset()
                
        s1 = s0.copy()
        t1 = t0        
        arr = []

        if self.is_rev_event:
            events = tuple(self.events)+(self.rev,)
            maxt = self.maxt
        else:
            events = tuple(self.events)
            maxt = self.rev
            
        detector = event_detector(self.model, events)
        
        if self.verbose: print('Montecarlo station-keeping:', end=' ')
        for i in range(N):
            up = self.rand_ball_one()*self.unc_pos_nd
            uv = self.rand_ball_one()*self.unc_vel_nd
            us = s1.copy()
            us[ :3] += up
            us[3:6] += uv
#            dv = self.correction.calc_dv(t1, us)
            if i == 0:
                # first correction == orbit determination
                dv = self.first_correction.calc_dv(t1, s1)
            else:
                dv = self.correction.calc_dv(t1, us)
            self.dvout.append([t1, *dv])        
            udv = self.rand_ball_percent(dv[3:6], self.unc_dv/100.)
            if i == 0:
                self.randout.append([*up, *uv, *dv[3:6]])
                s1 += dv
            else:
                self.randout.append([*up, *uv, *udv])
                s1[3:6] += udv
#            cur_rev, ev = event_detector(self.model, (self.rev,)).prop(s1, t1, t1+self.maxt, ret_df=False)
            cur_rev, ev = detector.prop(s1, t1, t1+maxt, ret_df=False)
            if ev.shape[0] > 0:
                self.evout.append(ev)

            t1 = cur_rev[-1,0]
            s1 = cur_rev[-1,1:].copy()
#            t1 = ev[-1,2]
#            s1 = ev[-1,3:].copy()
            arr.append(cur_rev.copy() if i==N-1 else cur_rev[:-1].copy())
            if self.verbose: print(i, end=' ')
        
        
        if self.verbose: print()
        arr = np.vstack(tuple(arr))
        if len(self.evout) > 0:
            self.evout = np.vstack(tuple(self.evout))
        
        if ret_df:
            df = self.model.to_df(arr)
            return df
        else:
            return arr



# -*- coding: utf-8 -*-
"""
Created on Sat Nov 24 00:31:13 2018

@author: stasb

This module is intended for event detection technique.

Event defined as time when some function E(t,s) called event-function reaches
specific value E*. In other words event detection is a root-finding process
for EF(t,s)=E(t,s)-E* function.
Detection splits into two steps: root separation and root refinement.
Root separation is finding two consecutive times ti, tj and states si, sj
(calculated through integration) such as EF(ti,si)*EF(tj,sj) < 0.
Root refinement is calculation of time t* and state s* such as 
|EF(t*,s*)| < eps, where eps is specified accuracy.

Root separation takes into account direction at which event-function reaches
its E* value:
    
direction <=> condition
    -1        EF(ti,si)>0 and EF(tj,sj)<0
     1        EF(ti,si)<0 and EF(tj,sj)>0
     0        EF(ti,si)*EF(tj,sj) < 0
     
Events can be terminal and non terminal. Terminal events tells integrator
when it should terminate integration process and therefore can be treated
as boundary conditions.


"""

import math
import pandas as pd
import numpy as np
from scipy.optimize import root as scipy_root
#from scipy.optimize import brentq
from scipy.interpolate import InterpolatedUnivariateSpline, interp1d
from numba import compiler, types
import pkg_resources

from .solout import default_solout, event_solout
from .models import base_model
from .plotter import plottable

class event_detector:
    '''
    Class event_detector is intended for event detection during
    integration of ODE system.
    '''
    columns = ['e', 'cnt', 'trm']
    
    def __init__(self, model, events, tol=1e-12, accurate_model=None):
#        if not issubclass(model.__class__, base_model):
        if not isinstance(model, base_model):
            raise TypeError("model should be instance of base_model class or of a subclass thereof")
        self.model = model
        if accurate_model is not None and not isinstance(accurate_model, base_model):
            raise TypeError("accurate_model should be instance of base_model class or of a subclass thereof")
        self.accurate_model = model if accurate_model is None else accurate_model
#        if not events:
#            raise ValueError("Empty events list")
        self.events = events
        self.solout = event_solout(self.events)
        self.tol = tol
        
    def prop(self, s0, t0, t1, 
             ret_df=True, 
             last_state='none'):
        '''
        
        Parameters
        ----------
        
        last_state : str
            Make last state of trajectory consistent with 'last' terminal event
                - 'none' - do not change trajectory;
                - 'last' - 'last' event is terminal event with greater row index in event states DataFrame;
                - 'mint' - 'last' event is terminal event that occured earlier than other terminal events
        '''
        for e in self.events:
            if hasattr(e, 'reset'):
                e.reset()
        old_solout, old_solout2 = self.model.solout, self.accurate_model.solout
        self.model.solout = self.solout
        df = self.model.prop(s0, t0, t1, ret_df=False)
        evdf = self.accurate(ret_df=False)
        self.model.solout, self.accurate_model.solout = old_solout, old_solout2
        
        # make last state consistent with terminal events
        last_state = last_state.lower()
        if evdf.shape[0] > 0 and last_state != 'none':
            cev, arr = self.split_data(evdf)
            arr_trm = arr[cev[:,2]==1]
            if arr_trm.shape[0] > 0:
                if last_state == 'last':
                    df[-1] = arr_trm[-1]
                elif last_state == 'mint':
                    s = arr_trm[np.argmin(arr_trm[:,0])]
                    df[-1] = s

        if ret_df:
            df = self.model.to_df(df)
            evdf = self.to_df(evdf)            
        return df, evdf
    
    def accurate(self, ret_df=True):
        self.accurate_model.solout = default_solout()
        
        evout = []
        for e in self.solout.evout:
            event = self.events[e[0]]
            if event.accurate:
                ts = self.solout.out[e[2]]
#                print('ts shape:', len(ts))
                t0 = ts[0]
                s0 = ts[1:]
                t1 = self.solout.out[e[2]+1][0]
#                print('root call:', 't0', t0, 't1', t1, 's0', s0)
                sa = self.root(event, t0, t1, s0)
#                print('acc shape:', len(sa))
            else:
                sa = self.solout.out[e[2]]
#                print('not acc shape:', len(sa))
            evout.append([e[0], e[1], event.terminal, *sa])
        
        if ret_df:
            df = self.to_df(evout)
            return df
        return np.array(evout)
        
    def root(self, event, t0, t1, s0):
        s_opt = [0] # for saving calculated state during root calculation
#        print('root finding', type(event))
        def froot(t, s0, t0):
            t = t[0] # because scipy.root passes [t] instead of t
            if t == t0:
                s = np.array([t, *s0])
            else:
                s = self.accurate_model.prop(s0, t0, t, ret_df=False)[-1]
            s_opt[0] = s
            res = event(t, s[1:])
#            print('t0', t0, 't', t, 's0', s0, 's', s, 'res', res)
#            print(event, t, s[1:], '->', res)
            return res
            
        scipy_root(froot, args=(s0, t0), x0=t0, tol=self.tol)
#        print('end of root finding')

#       scipy.optimize.solve_ivp uses:
#        brentq(froot, t0, t1, args=(s0, t0), xtol=self.tol, rtol=self.tol)
        return s_opt[0]
    
    def to_df(self, arr, columns=None):#, index_col=None):
#        if index_col is None:
#            index_col = event_detector.columns[0]
        if columns is None:
            columns = event_detector.columns+self.model.columns
        return self.model.to_df(arr, columns)#, index_col)
    
    def split_data(self, data):
        '''
        Splits data by columns: ['e', 'cnt', 'trm'] and ['t', 'x', 'y' ,...]
        '''
        if isinstance(data, pd.DataFrame):
#            d = data.reset_index()
            return data[event_detector.columns], data[self.model.columns]
        n = len(event_detector.columns)
        return data[:,:n], data[:,n:]

class base_event(plottable):
    '''
    Class base_event is a common interface for all event classes in OrbiPy.
    Event stores necessary data and calculates value of event-function which
    it represents. Event detection is a root finding of event-function.
    '''
    coord = 'e'
    
    def __init__(self, 
                 value=0,
                 direction=0,
                 terminal=True,
                 accurate=True,
                 count=-1):
        self.value = value
        self.terminal = terminal
        self.direction=direction
        self.accurate=accurate
        self.count=count
        
    def __call__(self, t, s):
        return 0
    
    def get_df(self):
        return pd.DataFrame({self.coord:self.value}, index=[0])
    
    def plot_df(self, df, ax, projection, **kwargs):
        p = projection.split('-')
        c = self.coord
        kwargs['label'] = self.__repr__()
        if p[0] == c:
            ax.axvline(df[c].values, **kwargs)
        elif p[1] == c:
            ax.axhline(df[c].values, **kwargs)
    
    def __repr__(self):
        return self.__class__.__name__ + \
               ':[val]%r [dir]%r [trm]%r [acc]%r [cnt]%r'%(self.value,
               self.direction, self.terminal, self.accurate, self.count)

class event_combine(base_event):
    '''
    Class event_combine combines list of events in one that looks
    like first event in list but acts like all of them simultaneously.
    This event occur when any of events in list occur.
    '''
    def __init__(self, events):
        if not events:
            raise ValueError('At least one event should be specified')
        self.events = events
            
    def __getattr__(self, attr):
#        print('combine_getattr', attr)
        return getattr(self.events[0], attr)

    def __call__(self, t, s):
        ret = 1.0
        for e in self.events:
            ret *= e(t, s)
        return ret
    
    def __repr__(self):
        return 'event_combine: ' + self.events.__repr__()
    
class event_chain(base_event):
    '''
    Class event_chain looks like last event in chain but works like sequence
    of events: events should occur in specified order and only last event
    will behave like event.
    All events in chain should be terminal!
    '''
    
    def __init__(self, events, autoreset=False):
        if not events:
            raise ValueError('At least one event should be specified')
        self.events = events
        self.autoreset = autoreset
        self.last = len(self.events) - 1
        self.select_event(0)

    def select_event(self, idx):
#        print('event_chain idx:', idx)
        self.idx = idx
        self.event_checker = event_solout([self.events[self.idx]])
            
        
    def reset(self):
        self.select_event(0)
        
    def __getattr__(self, attr):
#        print('chain_getattr', attr)
        return getattr(self.events[self.last], attr)

    def __call__(self, t, s):
        if self.idx == self.last:
            ret = self.events[self.idx](t, s)
            if self.autoreset and \
               self.event_checker(t, s) == -1:
                self.reset()
            return ret
        else:
            if self.event_checker(t, s) == -1:
                self.select_event(self.idx+1) # select next event
             # returning 0 will work because of strict inequalities in
             # event_solout
            return 0.0

    def __repr__(self):
        return 'event_chain: '+self.events.__repr__()

    
class center_event(base_event):
    '''
    Class center_event is a base class for all events that uses
    center (point) in calculations.
    '''
    def __init__(self, 
                 center=np.zeros(3), 
                 value=0, 
                 direction=0, 
                 terminal=True, 
                 accurate=True, 
                 count=-1):
        super().__init__(value, direction, terminal, accurate, count)
        self.center = center
        
    def __repr__(self):
        return super().__repr__() + ' [center]%r'%self.center

class center_angle_event(center_event):
    '''
    Class center_angle_event is a base class for all events that uses
    center (point) and angle in calculations.
    '''
    def __init__(self, 
                 center=np.zeros(3),
                 flip=False,
                 value=0,
                 direction=0, 
                 terminal=True, 
                 accurate=True, 
                 count=-1):
        '''
        Parameters
        ----------
        
        value : float
            Angle that belongs to [-pi, pi] segment.
        
        flip : bool
            Flip angle symmetrical to Y-axis therefore
            break at -pi/pi will be to the right of Y-axis.
        '''
        super().__init__(center, value, direction, terminal, accurate, count)
        self.flip = flip
                
    def __repr__(self):
        return super().__repr__() + ' [left]%r'%self.left
        
class model_event(base_event):
    '''
    Class model_event is a base class for all events that uses
    model in calculations.
    '''
    def __init__(self, 
                 model, 
                 value=0, 
                 direction=0, 
                 terminal=True, 
                 accurate=True, 
                 count=-1):
        super().__init__(value, direction, terminal, accurate, count)
        self.model = model

    def __repr__(self):
        return super().__repr__() + ' [model]%r'%self.model
            
class eventT(base_event):
    coord = 't'

    def __call__(self, t, s):
        return t - self.value

class eventSinT(base_event):      
    def __call__(self, t, s):
        return math.sin((t/self.value)*math.pi)

class eventX(base_event):      
    coord = 'x'
    
    def __call__(self, t, s):
        return s[0] - self.value
    
    def to_code(self, i):
        return 'val%03d = x-(%.18f)'%(i, self.value)
        
class eventY(base_event):        
    coord = 'y'

    def __call__(self, t, s):
        return s[1] - self.value

    def to_code(self, i):
        return 'val%03d = y-(%.18f)'%(i, self.value)
    
class eventZ(base_event):
    coord = 'z'

    def __call__(self, t, s):
        return s[2] - self.value

    def to_code(self, i):
        return 'val%03d = z-(%.18f)'%(i, self.value)
    
class eventVX(base_event):
    coord = 'vx'

    def __call__(self, t, s):
        return s[3] - self.value
    
class eventVY(base_event):
    coord = 'vy'

    def __call__(self, t, s):
        return s[4] - self.value
    
class eventVZ(base_event):
    coord = 'vz'

    def __call__(self, t, s):
        return s[5] - self.value
    
class eventAX(model_event):        
    coord = 'ax'

    def __call__(self, t, s):
        return self.model.right_part(t, s, self.model.constants)[3] - self.value
    
class eventAY(model_event):       
    coord = 'ay'

    def __call__(self, t, s):
        return self.model.right_part(t, s, self.model.constants)[4] - self.value
    
class eventAZ(model_event):
    coord = 'az'

    def __call__(self, t, s):
        return self.model.right_part(t, s, self.model.constants)[5] - self.value

class eventR(center_event):
    splits = 64
    def __call__(self, t, s):
        return ((s[0]-self.center[0])**2+
                (s[1]-self.center[1])**2+
                (s[2]-self.center[2])**2) - self.value**2
                
    def get_df(self):
        alpha = np.linspace(0, 2*np.pi, self.splits)
        c01 = self.center[0]+self.value*np.cos(alpha)
        c10 = self.center[1]+self.value*np.sin(alpha)
        c02 = self.center[0]+self.value*np.cos(alpha)
        c20 = self.center[2]+self.value*np.sin(alpha)
        c12 = self.center[1]+self.value*np.cos(alpha)
        c21 = self.center[2]+self.value*np.sin(alpha)
        z = np.zeros(self.splits, dtype=float)
        return pd.DataFrame({'x':np.hstack((c01,c02,  z)), 
                             'y':np.hstack((c10,  z,c12)),
                             'z':np.hstack((  z,c20,c21))})
    
    def plot_df(self, df, ax, projection, **kwargs):
        p = projection.split('-')
        all_prj = ('x-y', 'x-z', 'y-z')
        for i, prj in enumerate(all_prj):
            if projection in (prj, prj[::-1]):
                s = slice(i*self.splits,(i+1)*self.splits)
                ax.plot(df[p[0]][s].values, 
                        df[p[1]][s].values, **kwargs)
                break


class eventDR(center_event):
    def __call__(self, t, s):
        v = s[3:]   
        r = s[:3] - self.center
        r = r / (r[0]**2+r[1]**2+r[2]**2)**0.5
        return r[0]*v[0]+r[1]*v[1]+r[2]*v[2]

class eventRdotV(center_event):       
    def __call__(self, t, s):
        r = s[:3] - self.center
        v = s[3:6]
        return r[0]*v[0]+r[1]*v[1]+r[2]*v[2]
    
    def to_code(self, i):
        return '''val%03d = (x-%.18f)*vx+(y-%.18f)*vy+(z-%.18f)*vz'''%(i, *self.center)


class eventAlphaX(center_angle_event):
    def __call__(self, t, s):
        x, y = s[0]-self.center, s[1]
        if self.flip:
            x = -x
        angle = math.degrees(math.atan2(y, x))
        return angle - self.value

class eventOmegaX(center_event):
    def __call__(self, t, s):
        v = s[3:5]
        r = (s[0]-self.center, s[1])
        omega = (r[0]*v[1]-r[1]*v[0])/(r[0]**2+r[1]**2)
        return omega - self.value
    
class eventHyperboloidX(center_event):
    '''
    Circular hyperboloid oriented along X axis.
    XY-section is hyperbola defined by implicit equation:
        x^2/a^2-y^2/b^2 = 1
    '''
    param_t = 1.0
    splits = 64
    
    def __init__(self, a, b,
                 flip=False,
                 center=0.0, 
                 value=0, 
                 direction=0, 
                 terminal=True, 
                 accurate=True, 
                 count=-1):
        super().__init__(center, 0, direction, terminal, accurate, count)
        self.flip = flip
        self.a = a
        self.b = b

    def __call__(self, t, s):
        x, y = s[0]-self.center, (s[1]**2+s[2]**2)**0.5
        if self.flip:
            x = -x
        return x - self.a*(1 + (y/self.b)**2)**0.5

    def get_df(self):
        pt = np.linspace(-self.param_t, 
                          self.param_t, 
                          self.splits+1)
        x = self.a*np.cosh(pt) + self.center
        if self.flip:
            x = -x
        y = self.b*np.sinh(pt)
        return pd.DataFrame({'x':x, 'y':y})
    
    def plot_df(self, df, ax, projection, **kwargs):
#        p = projection.split('-')
#        all_prj = ('x-y', 'x-z', 'y-z')
        all_prj = ('x-y', 'x-z')
        for i, prj in enumerate(all_prj):
            if projection in (prj,):
                ax.plot(df['x'].values, 
                        df['y'].values, **kwargs)
#            if projection in (prj[::-1],):
#                ax.plot(df[p[1]].values, 
#                        df[p[0]].values, **kwargs)
                break

class eventParaboloidX(center_event):
    '''
    Circular paraboloid oriented along X axis.
    XY-section is parabola defined by explicit equation:
        x = y^2/a
    '''
    splits = 64
    
    def __init__(self, a,
                 flip=False,
                 center=0.0, 
                 value=0, 
                 direction=0, 
                 terminal=True, 
                 accurate=True, 
                 count=-1):
        super().__init__(center, 0, direction, terminal, accurate, count)
        self.flip = flip
        self.a = a

    def __call__(self, t, s):
        x, y = s[0]-self.center, (s[1]**2+s[2]**2)**0.5
        if self.flip:
            x = -x
        return x - y**2/self.a


class eventConeX(center_angle_event):
    def __call__(self, t, s):
        x, y = s[0]-self.center, (s[1]**2+s[2]**2)**0.5
        if self.flip:
            x = -x        
        angle = math.degrees(math.atan2(y, x))
        return angle - self.value

class eventInsidePathXY(center_angle_event):
    '''
    For detection of event when trajectory crosses 3D surface made by
    revolution of plane path around X-axis.
    '''
    def __init__(self, 
                 path, 
                 center,
                 flip=False,
                 splits=720,
                 direction=0, 
                 terminal=True, 
                 accurate=True, 
                 count=-1):
        super().__init__(center, flip, 0, direction, terminal, accurate, count)
#        if len(center) < 2:
#            raise TypeError('center should be iterable with 2 components (x, y)\n%r'%center)
        self.splits = splits
        self.set_path(path)
        
    def set_path(self, path):
        if not isinstance(path, np.ndarray) or path.ndim < 2:
            raise TypeError('path should be numpy array with 2 dimensions\n%r'%path)
        self.path = path
        x = self.path[:,0]-self.center
        if self.flip:
            x = -x
        y = self.path[:,1]
        theta = np.arctan2(y, x)
        order = np.argsort(theta)
        theta = theta[order]
        r = x**2 + y**2
        r = r[order]
#        self.rint = interp1d(theta[:-1], r[:-1], fill_value='extrapolate', kind='cubic')
        self.rint = InterpolatedUnivariateSpline(theta[:-1], r[:-1])
        # theta is uniformly distributed strictly increasing array
        self.theta = np.linspace(-np.pi, np.pi, self.splits)
        self.r = self.rint(self.theta)
        
    def theta_r(self, t, s):
        x, y, z = s[0]-self.center, s[1], s[2]
        if self.flip:
            x = -x
        r = x**2 + y**2 + z**2
        
#        if y < 0.0:
#            theta = math.atan2(-(y**2+z**2)**0.5, x)
#        else:
        theta = math.atan2((y**2+z**2)**0.5, x)
        return theta, r

    def interp(t, s, center, flip, x_arr, y_arr):
        '''
        Fast linear interpolation routine.
        x_arr should be uniformly distributed and sorted.
        '''
        # calculate r, theta
        x, y, z = s[0]-center, s[1], s[2]
        if flip:
            x = -x
            
        r = x**2 + y**2 + z**2
        
#        if y < 0.0:
#            theta = np.arctan2(-(y**2+z**2)**0.5, x)
#        else:
        theta = math.atan2( (y**2+z**2)**0.5, x)
        
        # linear interpolation part
        i = types.int64((theta-x_arr[0]) // (x_arr[1]-x_arr[0]))
#        i = math.floor((theta-x_arr[0]) / (x_arr[1]-x_arr[0]))
        r1 = y_arr[i] + (y_arr[i+1]-y_arr[i])/(x_arr[i+1]-x_arr[i])*(theta-x_arr[i])
#        print(t, s, r, r1, r-r1)
        return (r-r1)

    interp = \
    compiler.compile_isolated(interp, 
                              [types.double, types.double[:], types.double, 
                               types.boolean, types.double[:], types.double[:]],
                              return_type=types.double).entry_point
    
    def __call__(self, t, s):
        return eventInsidePathXY.interp(t, s, self.center, self.flip,
                                        self.theta, self.r)
    
    def deprecated__call__(self, t, s):
        theta, r = self.theta_r(t, s)
        r1 = self.rint(theta)
        print(r, r1)
        return r - r1  # inside path (event < 0)    
    
    def get_df(self):
        return pd.DataFrame({'x':self.path[:,0], 
                             'y':self.path[:,1]})
    
    def plot_df(self, df, ax, projection, **kwargs):
        p = projection.split('-')
        if projection in ('x-y','y-x'):
            ax.plot(df[p[0]].values, 
                    df[p[1]].values, **kwargs)


class eventSplitLyapunov(center_angle_event):
    '''
    For detection of event when trajectory crosses 3D surface made by
    revolution of plane Lyapunov orbit around X-axis.
    '''
    
    def __init__(self, 
                 lyapunov_orbit_half,
                 center,
                 flip=False,
#                 fname=pkg_resources.resource_filename(__name__, 'data/hlyapunov_sel1.csv'),
#                 orbit_idx=-500,
                 split_theta = 1.8461392981282345,
                 left = True,
                 splits=720,
                 direction=0, 
                 terminal=True, 
                 accurate=True, 
                 count=-1):
        super().__init__(center, flip, 0, direction, terminal, accurate, count)
        self.splits = splits
        self.split_theta = split_theta
        self.left = left
        self.set_orbit(lyapunov_orbit_half)
        
    def set_orbit(self, orbit):
        if not isinstance(orbit, np.ndarray) or orbit.ndim < 2:
            raise TypeError('orbit should be numpy array with 2 dimensions\n%r'%orbit)
        # orbit should be np.array, x->orbit[:,0], y->orbit[:,1]
        self.orbit = orbit.copy()
        mid_idx = self.orbit.shape[0]//2
        if self.orbit[mid_idx,1] < 0: # flip orbit
            self.orbit[:,1] = -self.orbit[:,1]
        x = self.orbit[:,0]-self.center
        if self.flip:
            x = -x
        y = self.orbit[:,1]
        theta = np.arctan2(y, x)
#        theta[theta < 0] += np.pi
        order = np.argsort(theta)
        theta = theta[order]
        r = x**2 + y**2
        r = r[order]
        self.rint = InterpolatedUnivariateSpline(theta, r)
        # theta is uniformly distributed strictly increasing array
        self.theta = np.linspace(0.0, np.pi, self.splits)
        self.r = self.rint(self.theta)
        
    def theta_r(self, t, s):
        x, y, z = s[0]-self.center, s[1], s[2]
        if self.flip:
            x = -x
        r = x**2 + y**2 + z**2
        
        theta = math.atan2((y**2+z**2)**0.5, x)
        
        return theta, r
    
    def get_xy(self, theta):
        r = self.rint(theta)**0.5
        x = r*math.cos(theta)
        if self.flip:
            x = -x
        x += self.center
        y = r*math.sin(theta)
        return np.array([x, y])

    def interp(t, s, center, flip, x_arr, y_arr, theta_s, left):
        '''
        Fast linear interpolation routine.
        x_arr should be uniformly distributed and sorted.
        '''
        # calculate r, theta
        x, y, z = s[0]-center, s[1], s[2]
        if flip:
            x = -x
        
        theta_xy = math.atan2(math.fabs(y), x)
#        print(theta_xy)
#        print('theta_xy, theta_s', theta_xy, theta_s)
        if (not left and (theta_xy > theta_s)) or (left and (theta_xy < theta_s)):
            return -1.0
        
        theta = math.atan2( (y**2+z**2)**0.5, x)
        
        r = x**2 + y**2 + z**2
        
        # linear interpolation part
        i = types.int64((theta-x_arr[0]) // (x_arr[1]-x_arr[0]))
#        i = math.floor((theta-x_arr[0]) / (x_arr[1]-x_arr[0]))
        r1 = y_arr[i] + (y_arr[i+1]-y_arr[i])/(x_arr[i+1]-x_arr[i])*(theta-x_arr[i])
#        print(t, s, r, r1, r-r1)
        return (r-r1)

    interp = \
    compiler.compile_isolated(interp, 
                              [types.double, types.double[:], types.double, 
                               types.boolean, types.double[:], types.double[:],
                               types.double, types.boolean],
                              return_type=types.double).entry_point
    
    def __call__(self, t, s):
        return eventSplitLyapunov.interp(t, s, self.center, self.flip,
                                         self.theta, self.r, self.split_theta,
                                         self.left)
    
    def deprecated__call__(self, t, s):
        theta, r = self.theta_r(t, s)
        theta_xy = math.atan2(math.fabs(s[1]), s[0])
        if (not self.left and (theta_xy > self.split_theta)) or (self.left and (theta_xy < self.split_theta)):
            return -1
        r1 = self.rint(theta)
        print(r, r1)
        return r - r1  # inside path (event < 0)    
    
    def get_df(self):
        return pd.DataFrame({'x':self.orbit[:,0], 
                             'y':self.orbit[:,1]})
    
    def plot_df(self, df, ax, projection, **kwargs):
        p = projection.split('-')
        if projection in ('x-y','y-x'):
            ax.plot(df[p[0]].values, 
                    df[p[1]].values, **kwargs)

    
class eventSPL(model_event):
    
    _data = {}
    _compiled_find_index = None
    _compiled_interp = None
    
    def __init__(self, 
                 model,
                 jc,
                 point='L1',
                 left=True,
#                 value=0, 
                 direction=0, 
                 terminal=True, 
                 accurate=True, 
                 count=-1):
        super().__init__(model, 0, direction, terminal, accurate, count)
        self.point = point
        self.left = left
        self.flip = (point == 'L1')
        self.center = model.__getattribute__(point)
        if eventSPL._compiled_find_index is None:
            eventSPL._compiled_find_index = \
                compiler.compile_isolated(eventSPL.find_index, 
                                          [types.double[:], types.double],
                                          return_type=types.int64).entry_point
        if eventSPL._compiled_interp is None:
            eventSPL._compiled_interp = \
                compiler.compile_isolated(eventSPL.interp,
                              [types.double, types.double[:], types.double, 
                               types.boolean, types.double[:], types.double[:],
                               types.double, types.boolean],
                              return_type=types.double).entry_point
        self.load_data(jc)        

    def find_index(arr, val):
        '''
        Find index of first element lower than val
        '''
        for i in range(len(arr)):
            if arr[i] < val:
                return i
        return -1

    def find_index_old(arr, val):
        mask = arr <= val
        return np.argmax(mask)
    
    def set_jc(self, jc):
        sp = eventSPL._data[self.point]['SP']
        pr = eventSPL._data[self.point]['PR']
        idx = eventSPL._compiled_find_index(sp[:,1], jc)
        if idx < 0:
            raise RuntimeError("Unable to find suitable orbit for Jacobi constant %f"%self.jc)
        self.split_theta = sp[idx, 2]
        self.selected_jc = sp[idx, 1]
        self.r = pr[idx].copy()
        self.jc = jc
        

    def load_data(self, jc):
        '''
        HLY = Horizontal Lyapunov Orbit Family
        PR = polar representation for HLY
        SP = Split Point
        '''
        if self.point in eventSPL._data:
            pr = eventSPL._data[self.point]['PR']
            sp = eventSPL._data[self.point]['SP']
        else:
            fname = 'HLY_' + self.point + '_' + self.model.const_set
            sp_fname = 'SP_' + fname + '.csv'
            pr_fname = 'PR_' + fname + '.npy'
            sp_path = pkg_resources.resource_filename(__name__, 'data/families/'+sp_fname)
            pr_path = pkg_resources.resource_filename(__name__, 'data/families/'+pr_fname)
            sp = np.loadtxt(sp_path)
            pr = np.load(pr_path)
            eventSPL._data[self.point] = {'SP': sp, 'PR': pr}

        self.set_jc(jc)
        self.theta = np.linspace(0., np.pi, 720)
        self.rint = InterpolatedUnivariateSpline(self.theta, self.r)

    def theta_r(self, t, s):
        x, y, z = s[0]-self.center, s[1], s[2]
        if self.flip:
            x = -x
        r = x**2 + y**2 + z**2
        
        theta = math.atan2((y**2+z**2)**0.5, x)
        
        return theta, r
    
    def get_xy(self, theta):
        r = self.rint(theta)**0.5
        x = r*math.cos(theta)
        if self.flip:
            x = -x
        x += self.center
        y = r*math.sin(theta)
        return np.array([x, y])

    def interp(t, s, center, flip, x_arr, y_arr, theta_s, left):
        '''
        Fast linear interpolation routine.
        x_arr should be uniformly distributed and sorted.
        '''
        # calculate r, theta
        x, y, z = s[0]-center, s[1], s[2]
        if flip:
            x = -x
        
        theta_xy = math.atan2(math.fabs(y), x)
#        print(theta_xy)
#        print('theta_xy, theta_s', theta_xy, theta_s)
        if (not left and (theta_xy > theta_s)) or (left and (theta_xy < theta_s)):
            return -1.0
        
        theta = math.atan2( (y**2+z**2)**0.5, x)
        
        r = x**2 + y**2 + z**2
        
        # linear interpolation part
        i = types.int64((theta-x_arr[0]) // (x_arr[1]-x_arr[0]))
#        i = math.floor((theta-x_arr[0]) / (x_arr[1]-x_arr[0]))
        r1 = y_arr[i] + (y_arr[i+1]-y_arr[i])/(x_arr[i+1]-x_arr[i])*(theta-x_arr[i])
#        print(t, s, r, r1, r-r1)
        return (r-r1)
    
    def __call__(self, t, s):
        return eventSPL._compiled_interp(t, s, self.center, self.flip,
                                         self.theta, self.r, self.split_theta,
                                         self.left)
        
        
  
class eventFOV(center_event):
    def __init__(self, 
                 orbit,
                 center, 
                 r, 
                 direction=0, 
                 terminal=True, 
                 accurate=True, 
                 count=-1):
        super().__init__(center, r, direction, terminal, accurate, count)
        self.set_orbit(orbit)

    def set_orbit(self, orbit):
        if not isinstance(orbit, np.ndarray) or orbit.ndim < 2 or orbit.shape[1] < 4:
            raise TypeError('[orbit] should be numpy array of (n, 4) shape: (t,x,y,z)\n%r'%orbit)
        self.orbit = orbit
        self.oint = interp1d(self.orbit[:,0], 
                             self.orbit[:,1:4],
                             axis=0,
                             kind='cubic',
                             fill_value='extrapolate')

    def __call__(self, t, s):
        cone = self.oint(t)[0]
        cone_c = self.center - cone
        cone_s =  s[:3] - cone
#        print(cone_c, cone_s)
#        try:
        d_c = (cone_c[0]**2+cone_c[1]**2+cone_c[2]**2)**0.5
        d_s = (cone_s[0]**2+cone_s[1]**2+cone_s[2]**2)**0.5
        alpha_cone_c = math.atan2(self.value,d_c)
        alpha_cone_s = math.acos(np.dot(cone_c, cone_s)/(d_c*d_s))
#        except:
#            pass
        return alpha_cone_c-alpha_cone_s
        


# -*- coding: utf-8 -*-
"""
Created on Tue Jul  2 10:14:44 2019

@author: stasb
"""
import math
import numpy as np
from scipy.interpolate import interp1d
from scipy.optimize import fminbound
from .interpolators import base_interpolator, linear_interpolator, ABab_mapper
    
class base_line:
    def __init__(self, interp=None):
        if interp is not None:
            if not isinstance(interp, base_interpolator):
                raise TypeError('This object is not base_interpolator inherited: %r'%interp)
        self.interp = interp
    
    def __call__(self, l):
        pass

    def get_points(self, n):
        if self.interp is not None:
            return self(self.interp.AB(n))
        return self(np.linspace(0.0, 1.0, n))
        
    def translate(self, v):
        return self
    
    def scale(self, s):
        return self

    def scale_at_center(self, s, center):
        return self.translate(center).scale(s).translate(-center)
    
    def get_closest_point(self, p, ret_l=False):
        l1 = fminbound(lambda l:np.sum((self(l)-p)**2), 0., 1.)
        if ret_l:
            return self(l1), l1
        return self(l1)
    
class line_segment(base_line):
    
    def __init__(self, x0, x1, interp=None):
        super().__init__(interp)
        self.x0 = x0
        self.x1 = x1
        
    def __call__(self, l):
        if self.interp is not None:
            li = self.interp(l)
        else:
            li = l
        
        if isinstance(li, np.ndarray):
            return self.x0*(1.0-li[:,None])+self.x1*li[:,None]
        
        return self.x0*(1.0-li)+self.x1*li    

    def translate(self, v):
        self.x0 += v
        self.x1 += v
        return self
    
    def scale(self, s):
        self.x0 *= s
        self.x1 *= s
        return self            
            

def heading_2d(x1, x2):
    '''
    Calculate angle from point x1 to point x2.
    '''
    dx = x2 - x1
    return math.degrees(math.atan2(dx[1], dx[0]))

def vector_2d(head):
    '''
    Calculate unit vector in head direction.
    '''    
    if isinstance(head, np.ndarray):
        rad = np.radians(head)
        return np.vstack((np.cos(rad),np.sin(rad))).T

    rad = math.radians(head)
    return np.array([math.cos(rad), math.sin(rad)])

class arc_segment(base_line):
    def __init__(self, center, head, fov, r, interp=None):
        super().__init__(interp)
        self.center = center
        self.head = head
        self.fov = fov
        self.r = r
        self.i1d = ABab_mapper(linear_interpolator(), A=0.0, B=1.0, 
                                       a=self.head-self.fov*0.5, 
                                       b=self.head+self.fov*0.5)

    def __call__(self, l):
        if self.interp is not None:
            li = self.interp(l)
        else:
            li = l
        
        heads = self.i1d(li)

        if isinstance(li, np.ndarray):
            return self.center[None,:] + self.r*vector_2d(heads)

        return self.center + self.r*vector_2d(heads)

    def translate(self, v):
        self.center += v
        return self
    
    def scale(self, s):
        self.center *= s
        self.r *= s
        return self            
    
class spline(base_line):
    
    def __init__(self, pts, kind='linear', interp=None):
        super().__init__(interp)
        self.pts = pts
        self.kind = kind
        self.updated = False
        self._update()
        
    def _update(self):
        self.l = np.concatenate(([0.], np.sum((self.pts[1:]-self.pts[:-1])**2,axis=1)**0.5))
        self.l = np.cumsum(self.l)/np.sum(self.l)
        self.i1d = interp1d(self.l, self.pts, kind=self.kind, 
                            axis=0, copy=False, 
                            fill_value='extrapolate')
        self.updated = True
        
    def __call__(self, l):
        if self.updated == False:
            self._update()
        if self.interp is not None:
            li = self.interp(l)
        else:
            li = l

        return self.i1d(li)
    
    def translate(self, v):
        self.pts += v[None, :]
        self.updated = False
        return self
    
    def scale(self, s):
        self.pts *= s
        self.updated = False
        return self
    
    @classmethod
    def linear(cls, pts, interp=None):
        return cls(pts, kind='linear', interp=interp)
    
    @classmethod
    def quad(cls, pts, interp=None):
        return cls(pts, kind='quadratic', interp=interp)

    @classmethod
    def cubic(cls, pts, interp=None):
        return cls(pts, kind='cubic', interp=interp)
    
if __name__ == '__main__':
    import matplotlib.pyplot as plt
    from .interpolators import ( log_interpolator, 
                                chain_interpolator,
                                flipX_interpolator,
                                flipY_interpolator)

    A, B = 1., 2.
    a, b = 0., 1.
    
    intrp = ABab_mapper(linear_interpolator(), A, B, a, b)
    intrp1 = flipX_interpolator(ABab_mapper(log_interpolator(), A, B, a, b))
    intrp2 = flipY_interpolator(ABab_mapper(log_interpolator(), a, b, A, B))
    
    intrp = chain_interpolator([intrp1, intrp2])
    
    x = np.linspace(A, B, 100)
    y = intrp(x)
    
    plt.figure(figsize=(10,10))
    plt.plot(x, y, '.-')
    
    x0 = np.array([10.,10.])
    x1 = np.array([20.,20.])
    ls = line_segment(x0, x1, log_interpolator(N=10))
    arc = arc_segment(x0, 45., 90., 10., log_interpolator(N=10))
    
    pts = arc.get_points(10)
    #ls.get_points(10)
    #pts[:,1] = -pts[:,1] + np.random.rand(pts.shape[0])*2-1
    
    pwl = spline.linear(pts, flipY_interpolator(log_interpolator(N=10)))
    pts1 = arc.get_points(50)
    
    plt.figure(figsize=(10,10))
    plt.plot(pts[:,0], pts[:,1], '.-')
    plt.plot(pts1[:,0], pts1[:,1], '.r')
    plt.axis('equal')
    
    


# -*- coding: utf-8 -*-
"""
Created on Mon Dec 31 00:39:18 2018

@author: stasb
"""
    
class iteration_printer:
    def __init__(self, idx=0, show_iter=True, sep=' '):
        self.idx = idx
        self.show_iter = show_iter
        self.sep = sep
        
    def __call__(self, i, count, arg):
        if self.show_iter:
            print('[%r/%r]'%(i, count), end=self.sep)
        if hasattr(self.idx, '__iter__'):
            print(*[arg[j] for j in self.idx], sep=self.sep)
        else:
            print(arg[self.idx])

class stepper:
    '''
    Stepper class is intended as convenient object-oriented analog of
    for loop. For each iteration (a.k.a. step) it calls user function
    func ('loop body') and passes to it current argument from specified
    range and reference to history where all previous steps are stored.
    '''
    def __init__(self, func, initial, args=(), prnt=iteration_printer()):
#        if not isinstance(func, base_func):
#            raise TypeError("Func should be an instance inheritor of base_func")
        self.func = func
        self.args = args
        self.prnt = prnt
        self.reset()
        if hasattr(initial, '__iter__'):
            for s in initial:
                self.history.append(s)
        else:
            self.history.append(initial)
    
    def reset(self):
        self.history = []
            
    def next_step(self, i):
        state = self.func(i, self.history, *self.args)
        if state is None:
            return False
        self.history.append(state)
        return True
    
    def make_steps(self, steps=100, verbose=True):
        if isinstance(steps, slice):
            r = range(steps.start, steps.stop, steps.step)
        elif hasattr(steps, '__iter__'):
            r = steps
        elif steps > 0:
            r = range(0, steps, 1)
        elif steps < 0:
            r = range(0, steps, -1)
        count = len(r)
        for idx, i in enumerate(r):
            if self.next_step(i):
                if verbose:
                    self.prnt(idx, count, self.history[-1])
#                    print('[%r/%r]'%(idx,count), '%r'%i, self.history[-1])
            else:
                if verbose:
                    print('Iteration stopped by function')
                break
#        if verbose:
#            print()
        return self



# -*- coding: utf-8 -*-
"""
Created on Wed Dec  5 18:22:53 2018

@author: stasb
"""

class mapper:
    '''
    mapper class is intended for conversion between physical value designation
    (typically presented as DataFrame column name) and categories such as:
    time(T), length(L), velocity(V) and other.
    '''

    columns = ('t', 'x', 'y', 'z', 'vx', 'vy', 'vz')
    indexes = tuple(range(len(columns)))
    categories = ('T', 'L', 'V')
    
    column2category = \
            {'t': 0,
             'x': 1,
             'y': 1,
             'z': 1,
             'vx':2,
             'vy':2,
             'vz':2,
             'dv':2
             }
            
    index2category = dict(enumerate(column2category.values()))    
    column2index = dict(zip(columns, indexes))
    
    def col2cat(column):
        '''
        Convert column (physical value designation) to category.
        
        Parameters
        ----------
        
        column : str
            Column name (physical value designation)
            
        Returns
        -------
        
        category : str
            Corresponding category name
        '''
        cat = mapper.column2category.get(column, None)
#        if cat is None:
#            raise RuntimeError("Can't find category for column name: %s"%column)
        return mapper.categories[cat] if cat is not None else None
    
    def idx2cat(index):
        return mapper.categories[mapper.index2category.get(index, None)]
    
    def col2idx(column):
        return mapper.column2index.get(column, None)
    
    def idx2col(index):
        return mapper.columns[index]


# -*- coding: utf-8 -*-
"""
Created on Tue Feb 12 17:52:14 2019

@author: stasb
"""

import numpy as np
import multiprocessing
import sys, os, time
from datetime import datetime, timedelta
import itertools

class mp:
    def __init__(self, workname, do_calc, do_save, folder=None):
        '''
        Parameters
        ----------
        
        workname : str
            Name of work to be done. Two files will be created/used:
                workname + '_done.csv' - jobs done list
                workname + '_todo.csv' - jobs todo list
                workname + '_fail.csv' - failed jobs list
            
        do_calc : function(job, folder)
            User function that made calculations
            
        do_save : function(item, folder)
            User function that save results
        
        folder : str
            Work folder
        '''
        self.workname = workname
        self.do_calc = do_calc
        self.do_save = do_save
        if folder in (None, ''):
            self.folder = os.getcwd()
        else:
            if not os.path.exists(folder):
                os.mkdir(folder)
            self.folder = folder
        self.done_path = os.path.join(self.folder, self.workname+'_done.csv')
        self.todo_path = os.path.join(self.folder, self.workname+'_todo.csv')
        self.fail_path = os.path.join(self.folder, self.workname+'_fail.csv')
    
    def run(self, p=2, run_async=False):
        '''
        Run parallel multiprocessed calculation.
        '''
        jobs = self.get_remaining_jobs()
        self.manager = multiprocessing.Manager()
        self.queue = self.manager.Queue()
        self.pool = multiprocessing.Pool(processes=p+1)
        
        mp.mprint('<<< pool of %d workers on %d jobs started >>>'%(p, len(jobs)))
        mp.mprint('<<< use [ctrl+c] to terminate processes >>>')
        
        self.pool.apply_async(mp.save_proxy, 
                              args=(len(jobs),
                                    self.do_save,
                                    self.queue, 
                                    self.done_path, 
                                    self.fail_path,
                                    self.folder))
        todo_jobs = [{'job':job,
                      'do_calc':self.do_calc,
                      'folder':self.folder,
                      'q':self.queue} for job in jobs]
        if run_async:
            self.pool.map_async(mp.calc_proxy, todo_jobs)
        else:
            self.pool.map(mp.calc_proxy, todo_jobs)
        self.pool.close()
        self.pool.join()

    def debug_run(self, job=None):
        '''
        Run calculation and saving for specified job in main process
        for debug and testing.
        '''        
        if job is None:
            job = self.get_remaining_jobs()[0]
            
        mp.mprint('<<< single process run on job %r >>>'%(job,))
        t0 = datetime.now()
        res = self.do_calc(job, self.folder)
        item = {'job':job, 'res':res}
        self.do_save(item, self.folder)
        dt = datetime.now()-t0
        mp.mprint_stat(dt, 0, 1)
        mp.mprint('<<< single process run finished >>>')

    def mprint(*args, **kwargs):
        '''
        Print args and flush buffer so result immediately goes to console
        '''
        print(*args, **kwargs)
        sys.stdout.flush()

    def dhms(seconds):
        '''
        Return the tuple of days, hours, minutes and seconds.
        '''
        minutes, seconds = divmod(seconds, 60)
        hours, minutes = divmod(minutes, 60)
        days, hours = divmod(hours, 24)
        return days, hours, minutes, seconds

    def generate_jobs(**kwdata):
        '''
        Make Cartesian product of iterables from kwdata
        '''
        return list(itertools.product(*kwdata.values()))
    
    def read_csv(path, sep=','):
        '''
        Read jobs from csv file.
        '''
        if not os.path.exists(path):
            return []
        with open(path, 'r') as f:
            lines = f.readlines()
        if len(lines) > 0:
            return list(map(lambda s: tuple(map(np.float64, s.strip().split(sep))), lines))
        return []
    
    def write_csv(path, jobs, sep=','):
        '''
        Save jobs to csv file.
        '''
        lines = [sep.join([str(arg) for arg in job])+'\n' for job in jobs]
        with open(path, 'wt') as f:
            f.writelines(lines)
            
    def append_csv(path, jobs, sep=','):
        '''
        Append jobs to csv file.
        '''
        lines = [sep.join([str(arg) for arg in job])+'\n' for job in jobs]
        with open(path, 'at') as f:
            f.writelines(lines)
    
    def update_todo_jobs(self, jobs):
        '''
        Main method to add jobs in TODO file.
        '''
        todo_jobs = set(mp.read_csv(self.todo_path))
        todo_jobs.update(set(jobs))
        mp.write_csv(self.todo_path, todo_jobs)
        return self

    def reset_done_jobs(self):
        '''
        When you need recalculate all jobs use this method to delete
        DONE file with done jobs.
        '''
        if os.path.exists(self.done_path):
            os.remove(self.done_path)
        return self

    def reset_failed_jobs(self):
        '''
        When you need recalculate all jobs use this method to delete
        FAIL file with failed jobs.
        '''
        if os.path.exists(self.fail_path):
            os.remove(self.fail_path)
        return self

    def get_remaining_jobs(self, exclude_failed=True):
        '''
        Uses TODO, DONE and FAILED files to retrieve remaining jobs. 
        '''
        todo_jobs = set(mp.read_csv(self.todo_path))
        if not todo_jobs:
            return []
        done_jobs = set(mp.read_csv(self.done_path))
        
        if exclude_failed:
            fail_jobs = set(mp.read_csv(self.fail_path))
            return list(todo_jobs.difference(done_jobs).difference(fail_jobs))
        else:
            return list(todo_jobs.difference(done_jobs))
        
    def get_failed_jobs(self):
        '''
        Uses FAILED file to retrieve failed jobs. 
        '''
        failed_jobs = mp.read_csv(self.failed_path)
        return failed_jobs

    def dhms_string(dt):
        '''
            dt : timedelta
        '''
        return '%02d:%02d:%02d:%05.2f'%mp.dhms(dt.total_seconds())
    
    def mprint_stat(dt, est_count, done_count):
        '''
        Print average time per job, estimated time of arrival,
        finish datetime.
        '''
        name = (' LAST %d '%done_count).center(32,'-')
        avg = dt.total_seconds()/done_count
        eta = timedelta(seconds=avg*est_count)
        fin = datetime.now()+eta
        
        mp.mprint(name+'\n',
                  '< AVG: %.2fs >\n'%avg,
                  '< ETA: %s >\n'%mp.dhms_string(eta),
                  '< FIN: %s >\n'%fin.strftime('%d-%b-%Y %H:%M:%S'),
                  '-'*32, sep='')
        
    def save_proxy(N, do_save, queue, done_path, fail_path, folder):
        '''
        This function acts in separate process and saves calculated 
        results from queue.
        '''
        t0 = datetime.now()
        t10 = t0
        i = 0
        while i < N:
            if queue.empty():
                time.sleep(0.1)
            else:
                item = queue.get()
                if item['res'] is not None:                    
                    mp.mprint('< SAVE #%d/%d >'%(i,N))
                    do_save(item, folder)
                    mp.append_csv(done_path,[item['job']])
                else:
                    mp.append_csv(fail_path,[item['job']])

                i += 1
                if i%10 == 0:
                    dt = datetime.now()-t0
                    mp.mprint_stat(dt,N-i,i)
                    dt10 = datetime.now()-t10
                    t10 = datetime.now()
                    if i > 10:
                        mp.mprint_stat(dt10,N-i,10)
        # finishing
        dt = datetime.now()-t0
        mp.mprint_stat(dt,N-i,i)
        mp.mprint('<<< pool finished working on %d jobs >>>'%(N))
        
    def calc_proxy(arg):
        '''
        This function is wrapper around do_calc user function.
        Runs in multiple processes, calls do_calc, work on exceptions.
        '''
        job = arg['job']
        do_calc = arg['do_calc']
        mp.mprint('< CALC:', job, '>')
        try:
            t0 = datetime.now()
            res = do_calc(job, arg['folder'])
        except BaseException as e:
            mp.mprint('< EXCEPTION: %r >'%e)
            res = None
        finally:
            t = datetime.now() - t0
            mp.mprint('< JOB %r DONE >\n< CPU: %.2fs >'%(job,t.total_seconds()))
            arg['q'].put({'job':job, 'res':res})
        
    def test_calc(job, folder):
        # actual calculation work will be done here!
        time.sleep(np.random.randint(1,5))
        if np.random.randint(1,4) > 2:
            raise RuntimeError('Calculation error')
        res = job[1]*job[0]
        return res
    
    def test_save(item, folder):
        # results will be saved here
        mp.mprint('Saving result: ', item['res'], 
                  'of job:', item['job'],
                  'into folder', folder)
        pass
    
if __name__ == '__main__':
    jobs = mp.generate_jobs(x = list(range(10)), y = list(range(10))[::-1])
    m = mp('mp_test_work', 
           do_calc=mp.test_calc,
           do_save=mp.test_save).update_todo_jobs(jobs).reset_fail_jobs()
    m.debug_run()
#    m.run(p=2)



# -*- coding: utf-8 -*-
"""
Created on Sun Dec  2 19:49:32 2018

@author: stasb
"""

#import numpy as np
import pandas as pd
from itertools import product
from .models import nondimensional_model
from .mapper import mapper

default_units_mapper = \
         {'T': { 's':1.,
                 'sec':1.,
                 'min':60.,
                 'h':60*60.,
                 'hour':60*60.,
                 'd':24*60*60.,
                 'day':24*60*60.,
                 'w':7*24*60*60.,
                 'week':7*24*60*60.,
                 'M':30*24*60*60.,
                 'month':30*24*60*60.,
                 'Y':365*24*60*60.,
                 'year':365*24*60*60.
               },

          'L': { 'mm':1e-6,
                 'cm':1e-5,
                 'm' :1e-3,
                 'km':1.,
                 'Mm':1e3,
                 'Gm':1e6,
                 'au':149597870.700,
                 'AU':149597870.700,
                 }
          }
          
class scaler:
    '''
    Class scaler is intended for conversion between physical units of several
    categories such as: time, length, velocity and other.
    By default it uses default_units_mapper defined in scaler.py module.
    '''
    def __init__(self, 
                 time_units=('s','s'),
                 length_units=('km','km'),
                 velocity_units=('km/s','km/s'),
                 units_mapper=default_units_mapper.copy()):
        '''
        Initializes scaler instance using specified mapper and units.
        
        Parameters
        ----------
        time_units, length_units, velocity_units : tuple(str, str)
            Default units used for data conversion in convert_data method by default.
            
        units_mapper : dict(str, dict(str, float))
            Mapping dict for each category and unit in category consists of
            coefficients used to convert values between physical units.
        
        Example
        -------
        import orbipy as op
        import numpy as np
        
        scale = op.scaler()
        
        # convert 10 mm to cm
        print(scale(10,'mm-cm'), 'cm')
        
        # convert 0,10,...,90 seconds to minutes
        print(*scale(np.arange(10)*10, 's-m'), sep=' m\n', end=' m\n')
        
        model = op.crtbp3_model()
        scale = op.scaler.from_model(model)
        
        # convert 2*pi nondimensional time units to days
        # for Sun-Earth system (365.0)
        print(scale(2*np.pi,'nd-d'))
        
        See also
        --------
        default_units_mapper from scaler module.
        
        '''
        self.units_mapper = units_mapper
        v_keys = ['%s/%s'%(l,t) for l, t in product(self.units_mapper['L'].keys(),  self.units_mapper['T'].keys())]
        v_vals = [          l/t for l, t in product(self.units_mapper['L'].values(),self.units_mapper['T'].values())]
        self.units_mapper['V'] = dict(zip(v_keys,v_vals))
        self.units = dict(zip(mapper.categories, (time_units, length_units, velocity_units)))
        for cat in self.units:
            if self.get_category(self.units[cat][0], self.units[cat][1]) is None:
                raise ValueError("Can't convert [%s] to [%s] units from [%s] category"%(*self.units[cat], cat))
    
    def get_category(self, from_units, to_units):
        '''
        Tries to get category of units which contains both: from_units and to_units.
        
        Parameters
        ----------
        from_units, to_units : str
            Physical units, like 'm', 'km/s', 'd', 'nd'.

        Return
        ------
        category : str or None
            Category of units (key from units_mapper dict) if both of specified units
            can be located in such category. None, otherwise.
        '''
        for c in self.units_mapper:
            if (from_units in self.units_mapper[c]) and (to_units in self.units_mapper[c]):
                return c
        return None
    
    @classmethod
    def from_model(cls, model,
                   time_units=('nd','d'),
                   length_units=('nd','km'),
                   velocity_units=('nd/nd','km/s'),
                   nondim='nd'):
        '''
        Constructor from_model initializes scaler class instance with additional
        nondimensional units gathered from model.
        
        Parameters
        ----------
        model : nondimensional_model instance
            Model from which nondimensional coefficients will be used.
            
        nondim : str
            Symbol for nondimensional units.
            
        rest : parameters
            See __init__.
        '''
        if not isinstance(model, nondimensional_model):
            raise TypeError('Only nondimensional models supported')
        umapper = default_units_mapper.copy()
        nd_coefs = model.get_nd_coefs()
        umapper['T'][nondim] = nd_coefs['T']
        umapper['L'][nondim] = nd_coefs['L']
        return cls(time_units, length_units, velocity_units, units_mapper=umapper)
        
    def get_coef(self, from_units='km', to_units='m'):
        '''
        Calculate conversion coefficient to convert value from from_units to to_units.
        
        Parameters
        ----------
        from_units, to_units : str
            Physical units, like 'm', 'km/s', 'd', 'nd'.
            
        Return
        ------
        coef : float
            Conversion coefficient.
        '''
        category = self.get_category(from_units, to_units)
        if not category:
            raise ValueError("Can't convert [%s] to [%s]"%(from_units, to_units))
        cat = self.units_mapper[category]
        return cat[from_units] / cat[to_units]
    
    def convert(self, value, from_units='km', to_units='m'):
        '''
        Converts value from from_units to to_units.
        
        Parameters
        ----------
        value : float
            Value specified in from_units physical units.
            
        from_units, to_units : str
            Physical units, like 'm', 'km/s', 'd', 'nd'.
            
        Return
        ------
        new_value : float
            Converted value.        
        '''
        return value * self.get_coef(from_units, to_units)
    
    def __call__(self, value, from_to='km-m'):
        '''
        Converts value from one physical units to another.
        Short version of convert method.
        
        Parameters
        ----------
        
        from_to : str
            Physical units for convertation in format: 'unit_from-unit_to'.
        '''
        return value * self.get_coef(*from_to.split('-'))
    
    def convert_data(self, data,
                     time_units=None,
                     length_units=None,
                     velocity_units=None):
        '''
        Converts data using specified or default units.
        
        Parameters
        ----------
        data : numpy.ndarray or pandas.DataFrame
            Data consists of columns that intended for conversion according to
            specified units.
            
        time_units, length_units, velocity_units : tuple(str, str)
            Units used for columns conversion. Each column corresponds to
            category of units (time, length, velocity) and will be converted
            from category_units[0] to category_units[1]. If specified units is
            None then default units will be used (self.units).
            
        Return
        ------
        new_data : same type as data
            Converted data.      
        '''
        if time_units is None:
            time_units = self.units['T']
        if length_units is None:
            length_units = self.units['L']
        if velocity_units is None:
            velocity_units = self.units['V']
        units = dict(zip(mapper.categories, (time_units, length_units, velocity_units)))
        df = data.copy()
        if isinstance(data, pd.DataFrame):
            for col in df.columns:
                cat = mapper.col2cat(col)
                if cat is not None:
                    u = units[cat]
                    df[col] = self.convert(df[col], from_units=u[0], to_units=u[1])
#            df.index = self.convert(df.index,
#                                    from_units=time_units[0],
#                                    to_units=time_units[1])
        else:
            for i in range(df.shape[1]):
                if i in mapper.indexes:
                    u = units[mapper.idx2cat(i)]
                    df[:,i] = self.convert(df[:,i], from_units=u[0], to_units=u[1])
            
        return df




# -*- coding: utf-8 -*-
"""
Created on Sun Jul  7 12:34:13 2019

@author: stasb
"""

import math
import numpy as np
#from scipy.interpolate import interp1d

class base_interpolator:
    '''
    Root class for all interpolators.
    Calculates mapping from [0, 1] to [0, 1]
    '''
            
    def __call__(self, x):
        return x

    def AB(self, n):
        return np.linspace(0.0, 1.0, n)
    
    def ab(self, n):
        return self(self.AB(n))
    
class ABab_mapper(base_interpolator):
    '''
    Calculates mapping from [A, B] to [a, b].
    interp should be [0,1]->[0,1] interpolator
    '''

    def __init__(self, interp, A=0.0, B=1.0, a=0.0, b=1.0):
        if not isinstance(interp, base_interpolator):
            raise TypeError('interp should be an instance of base_interpolator')
        self.d = B-A
        if math.fabs(self.d) < 1e-16:
            raise RuntimeError('A and B are very close, B-A=%f'%self.d)
        self.A = A
        self.B = B
        self.a = a
        self.b = b
        self.interp = interp
        
    def __call__(self, x):
        x_in = (x - self.A) / self.d
        x_out = self.interp(x_in)
        return self.a*(1-x_out)+self.b*x_out
    
    def AB(self, n):
        return np.linspace(self.A, self.B, n)
        
    def __getattr__(self, attr):
        return getattr(self.interp, attr)


class linear_interpolator(base_interpolator):        
    pass        
    
class log_interpolator(base_interpolator):
    def __init__(self, N=math.e):
        if N <= 0 or N == 1:
            raise ValueError("logarithm base should be >= 0.0 and not equal 1.0")
        self.iN = 1/N
        self.logN = math.log(N)
        self.x0 = self.iN / (1.0 - self.iN)
        self.y0 = -math.log(self.x0) / self.logN
        
    def __call__(self, x):
        return np.log(x+self.x0)/self.logN+self.y0

class flipY_interpolator(base_interpolator):
    def __init__(self, intrp):
        if not isinstance(intrp, base_interpolator):
            raise TypeError('This object is not base_interpolator inherited: %r'%intrp)
        self.intrp = intrp
        
    def __getattr__(self, attr):
        return getattr(self.intrp, attr)

    def __call__(self, x):
        return self.a+self.b-self.intrp(x)
    
class flipX_interpolator(base_interpolator):
    def __init__(self, intrp):
        if not isinstance(intrp, base_interpolator):
            raise TypeError('This object is not base_interpolator inherited: %r'%intrp)
        self.intrp = intrp
        
    def __getattr__(self, attr):
        return getattr(self.intrp, attr)
        
    def __call__(self, x):
        return self.intrp(1.0-x)

class chain_interpolator(base_interpolator):
    def __init__(self, interpolators):
        if len(interpolators) == 0:
            raise ValueError('No one interpolator passed')
        for intrp in interpolators:
            if not isinstance(intrp, base_interpolator):
                raise TypeError('This object is not base_interpolator inherited: %r'%intrp)
        self.interpolators = interpolators
        
    def __getattr__(self, attr):
        return getattr(self.interpolators[0], attr)

    def __call__(self, x):
        r = x
        for intrp in self.interpolators:
            r = intrp(r)
        return r



# -*- coding: utf-8 -*-
"""
Created on Fri Nov 23 00:23:47 2018

@author: stasb
"""

import numpy as np
import pandas as pd
import scipy.integrate
#from numba import compiler, types


class base_integrator:
    '''
    Base interface for all integrator wrappers that will be used in OrbiPy.
    Integrator should provide solout functionality similar to 
    scipy.integrate.ode.
    '''
    def __init__(self, params):
        self._solout = None
        self._params = params
        
    def get_params(self):
        return self._params
    
    def set_params(self, **params):
        self._params.update(params)
        
    @property
    def solout(self):
        '''Get/set solout callable.
           Solout callable used for full trajectory output and
           event detection.
        '''
        return self._solout
        
    @solout.setter
    def solout(self, solout):
        '''Set solout callable.'''
        self._solout = solout

    def integrate(self, func, s0, t0, t1):
        '''Integrate func from initial state s0 at time t0 to time t1.'''
        return np.array([t0, *s0])

class quasi_integrator(base_integrator):
    '''
    Quasi integrator provides integrator interface as a proxy for already 
    calculated trajectory data (numpy.array or pandas.DataFrame). Can be
    used for calculation of additional events on existing orbit.
    '''
    def __init__(self, data, solout=None):
        self.data = data
        self._solout = solout

    def get_params(self):
        return None
    
    def set_params(self, **params):
        pass
        
    @property
    def solout(self):
        return self._solout
        
    @solout.setter
    def solout(self, solout):
        self._solout = solout

    def integrate(self, func, s0, t0, t1, fargs=()):
        ''' Runs all states from self.data trough solout callable.
        '''
        if isinstance(self.data, pd.DataFrame):
            t = self.data.values[:,0]
            arr = self.data.values[:,1:]
        else:
            t = self.data[:,0]
            arr = self.data[:,1:]
        #self._solout.reset()
        for i in range(len(t)):
            if self._solout(t[i], arr[i]) == -1:
                break
        return self.data
        
class dop853_integrator(base_integrator):
    '''
    Adapter for scipy.integrate.ode class with dop853 integrator.
    '''
    def __init__(self,
                 params={'rtol':1e-12, 'atol':1e-12, 
                         'nsteps':100000, 'max_step':np.inf},
                 solout=None):
        if params and isinstance(params, dict):
            self._params = params.copy()
        else:
            self._params = {}
        self._solout = solout
        self.integrator = None
        self.func = None

    
    def set_params(self, **params):
        self._params.update(params)
        if self.integrator:
            self.integrator.set_integrator('dop853', **self._params)

    @property
    def solout(self):
        return self._solout
        
    @solout.setter
    def solout(self, solout):
        self._solout = solout
        if solout and self.integrator:
            self.integrator.set_solout(self._solout)

    def integrate(self, func, s0, t0, t1, fargs=()):
        if not func:
            raise ValueError('Function is not defined')
        
        if not self.integrator or func != self.func:
            self.func = func
            self.integrator = scipy.integrate.ode(func)
            self.integrator.set_integrator('dop853', **self._params)
                
#        print(self.params['max_step'])
        if self._solout:
            self.integrator.set_solout(self._solout)
        self.integrator.set_f_params(*fargs)        
        self.integrator.set_initial_value(s0, t0)
        return self.integrator.integrate(t1)
    
class dopri5_integrator(base_integrator):
    '''
    Adapter for scipy.integrate.ode class with dopri5 integrator.
    '''
    def __init__(self, 
                 params={'rtol':1e-12, 'atol':1e-12, 
                         'nsteps':100000, 'max_step':np.inf}, 
                 solout=None):
        if params and isinstance(params, dict):
            self._params = params.copy()
        else:
            self._params = {}
        self._solout = solout
        self.integrator = None
        self.func = None
    
    def set_params(self, **params):
        self._params.update(params)
        if self.integrator:
            self.integrator.set_integrator('dopri5', **self._params)

    @property
    def solout(self):
        return self._solout
        
    @solout.setter
    def solout(self, solout):
        self._solout = solout
        if solout and self.integrator:
            self.integrator.set_solout(self._solout)

    def integrate(self, func, s0, t0, t1, fargs=()):
        if not func:
            raise ValueError('Function is not defined')
        
        if not self.integrator or func != self.func:
            self.func = func
            self.integrator = scipy.integrate.ode(func)        
            self.integrator.set_integrator('dopri5', **self._params)
            
        if self._solout:
            self.integrator.set_solout(self._solout)
        self.integrator.set_f_params(*fargs)  
        self.integrator.set_initial_value(s0, t0)
        return self.integrator.integrate(t1)



# -*- coding: utf-8 -*-
"""
Created on Sat Jan 26 16:08:03 2019

@author: stasb
"""

import numpy as np
import math
from scipy.optimize import brent
from .events import event_detector, base_event
from .models import base_model

class base_direction:    
    '''
    Class direction is a base class for all direction calculation
    algorithms in OrbiPy. Direction is a normalized 3D vector.
    Directions are used in correction methods to define axis of
    correction maneuver.
    '''
    def __call__(self, t, s):
        return np.array([1., 0., 0.])
    
class xy_angle_direction(base_direction):
    '''
    Constant direction in XY plane determined by angle
    counted from X axis anticlockwise.
    '''
    def __init__(self, angle):
        self.angle = angle
        rads = math.radians(self.angle)
        self.direction = np.array([math.cos(rads), math.sin(rads), 0.])
        
    def __call__(self, t, s):
        return self.direction

class x_direction(base_direction):
    '''
    Constant direction along X axis.
    '''
    direction = np.array([1., 0., 0.])
        
    def __call__(self, t, s):
        return self.direction


class y_direction(base_direction):
    '''
    Constant direction along Y axis.
    '''
    direction = np.array([0., 1., 0.])

    def __call__(self, t, s):
        return self.direction

        
class velocity_direction(base_direction):
    '''
    Direction along spacecraft velocity.
    '''
    def __call__(self, t, s):
        return s[3:6]/(s[3]**2+s[4]**2+s[5]**2)**0.5

class unstable_direction_stm(base_direction):
    '''
    Unstable direction at which most effective corrections
    can be made. Calculated using eigenvector corresponding
    to real positive (i.e. unstable) eigenvalue of State
    Transition Matrix calculated after a specified time
    elapsed since initial time.
    '''
    def __init__(self, 
                 model,
                 event=np.pi,
                 maxt=10*np.pi,
                 ignore_z=False,
                 alg='dot'):
        if not isinstance(model, base_model):
            raise TypeError('model should be an instance of base_model')
        if not hasattr(model, 'stm') or model.stm==False:
            raise ValueError('model should support STM calculation')
        self.model = model
        self.maxt = maxt
        self.event = event
        self.is_event = isinstance(event, base_event)
        self.ignore_z = ignore_z
        self.alg = alg

    def __call__(self, t, s):
        s0 = self.model.get_zero_state()
        s0[:s.shape[0]] = s
        if self.is_event:
            detector = event_detector(self.model, [self.event])
            _, ev = detector.prop(s0, t, t+self.maxt)
            _, df = detector.split_data(ev)            
            stm = df.values[-1,7:].reshape(6,6)
        else:
            arr = self.model.prop(s0, t, t+self.event, ret_df=False)
            # t(0),x(1),y(2),z(3),vx(4),vy(5),vz(6),stm00(7), ...
            stm = arr[-1,7:].reshape(6,6)
            
        if self.alg == 'dot':
            eigvals, eigvecs = np.linalg.eig(stm[0:3,3:6].T@stm[0:3,3:6])
        else: # 'inv'
            eigvals, eigvecs = np.linalg.eig(np.linalg.inv(stm)[0:3,0:3])

        idx = 0 if eigvals[0].real > eigvals[1].real else 1
        
        d = eigvecs[:,idx].real

        if d[0] < 0.:
            d = -d
        if self.ignore_z:
            d[2] = 0.
        d = d / (d[0]**2+d[1]**2+d[2]**2)**0.5
        return d
    
    
class unstable_direction(base_direction):
    '''
    Unstable direction at which most effective corrections
    can be made. Calculated by maximization of velocity deviation
    after forward propagation of disturbed state for maxt time.
    '''
    def __init__(self, 
                 model,
                 disturbance=1e-8,
                 maxt=np.pi,
                 tol=1e-5,
                 iterations=100):
        if not isinstance(model, base_model):
            raise TypeError('model should be an instance of base_model')
        self.model = model
        self.disturbance = disturbance
        self.maxt = maxt
        self.iterations = iterations
        self.tol = tol

    def __call__(self, t, s):
        s0 = self.model.get_zero_state()
        s0 = s[:s0.shape[0]]
        # t(0),x(1),y(2),z(3),vx(4),vy(5),vz(6),stm00(7), ...
        c = self.model.prop(s0, t, t+self.maxt, ret_df=False)[-1]
        def goal(beta):
            s1 = s0.copy()
            s1[3] += self.disturbance * math.cos(beta)
            s1[4] += self.disturbance * math.sin(beta)
            p = self.model.prop(s1, t, t+self.maxt, ret_df=False)[-1]
            return -((p[0]-c[0])**2+(p[1]-c[1])**2)
        
        ret = brent(goal, brack=(0,2*np.pi), maxiter=self.iterations, 
                    tol=self.tol, full_output=True)
#        print(ret)
        
        beta = ret[0]
        d = np.array([math.cos(beta), math.sin(beta), 0.])
        
        if d[0] < 0.:
            d = -d

        return d



# -*- coding: utf-8 -*-
"""
Created on Fri Apr 26 11:31:16 2019

@author: stasb
"""

from .models import crtbp3_custom_model
from numba import compiler, types
import numpy as np # do not remove - crtbp_ode uses it
import math # do not remove - crtbp_ode uses it
import warnings

class base_model_builder:
    pass

class crtbp3_model_builder(base_model_builder):
    '''
    
    Example
    -------
    
    model_events = [op.eventVX(), op.eventY()]

    mb = op.crtbp3_model_builder(stm=True, events=model_events,
                                 a=1e-6, c=5e-4)
    
    # This model will be sensitive to vx==0 and y=0 events
    # and generate states frequently near this events.
    cmodel = mb.buid_model()

    '''
    
    
    _header_code = \
'''
def crtbp_ode(t, s, constants):
    
    # initialization and pre calculations
    
    mu2 = constants[0]
    mu1 = 1 - mu2
    
    x = s[0]
    y = s[1]
    z = s[2]
    vx = s[3]
    vy = s[4]
    vz = s[5]
    
    ds = np.empty(%d)
    
    yz2 = y * y + z * z;
    xmu2 = x + mu2
    xmu1 = x - mu1

    r1 = (xmu2**2 + yz2)**0.5
    r2 = (xmu1**2 + yz2)**0.5
    r13, r15 = r1**(-3), r1**(-5)
    r23, r25 = r2**(-3), r2**(-5)
'''

    _main_code = \
'''
    # main ode calculation
    
    mu12r12 = (mu1 * r13 + mu2 * r23);

    ax =  2. * vy + x - (mu1 * xmu2 * r13 + mu2 * xmu1 * r23);
    ay = -2. * vx + y - mu12r12 * y;
    az =              - mu12r12 * z;

    ds[0] = vx
    ds[1] = vy
    ds[2] = vz
    ds[3] = ax
    ds[4] = ay
    ds[5] = az
'''        
    _stm_code =\
'''
    # stm calculation
    
    stm0 = np.ascontiguousarray(s[6:42]).reshape(6,6)

    Uxx = 1. - mu12r12 + 3*mu1*xmu2**2*r15 + 3*mu2*xmu1**2*r25
    Uxy = 3*mu1*xmu2*y*r15 + 3*mu2*xmu1*y*r25
    Uxz = 3*mu1*xmu2*z*r15 + 3*mu2*xmu1*z*r25
    Uyy = 1. - mu12r12 + 3*mu1*y**2*r15 + 3*mu2*y**2*r25
    Uyz = 3*mu1*y*z*r15 + 3*mu2*y*z*r25
    Uzz = -mu12r12 + 3*mu1*z**2*r15 + 3*mu2*z**2*r25
    
    A = np.array(((0.,  0.,  0.,   1., 0., 0.),
                  (0.,  0.,  0.,   0., 1., 0.),
                  (0.,  0.,  0.,   0., 0., 1.),
                  (Uxx, Uxy, Uxz,  0., 2., 0.),
                  (Uxy, Uyy, Uyz, -2., 0., 0.),
                  (Uxz, Uyz, Uzz,  0., 0., 0.)))

    stm1 = np.dot(A,stm0)
    ds[6:42] = stm1.ravel()
'''
    _event_header =\
'''
    # event perception
    
    def impulse(x, a, c, d):
        c2 = 2*c**2
        return -a*d*x*math.exp(-math.fabs(x)**d/c2)/c2
'''

    _event_code =\
'''
    %s
    ds[%d] = impulse(val%03d, %e, %e, %e)
'''

    _unsupported_event_code =\
'''
    ds[%d] = 0.0
'''
    
    _footer_code = \
'''
    return ds    
''' 
    
    def __init__(self, stm=False, events=[], 
                 a=1e-5, c=1e-5, d=1.5):
        
        self.events = events
        self.stm = stm
        
        state_size = 6 + (36 if stm else 0) + len(events)
        
        self.code = '' + self._header_code%state_size + self._main_code
        
        if stm:
            self.code += self._stm_code
        
        # events
        if events:
            self.code += self._event_header
            
            for i, event in enumerate(events):
                if type(event) == dict:
                    e = event['event']
                    a1, c1, d1 = event.get('a', a), event.get('c', c), event.get('d', d)
                else:
                    e = event
                    a1, c1, d1 = a, c, d

                if not hasattr(e, 'to_code'):
                    warnings.warn("Event %r does not have to_code method"%e)
                    self.code += self._unsupported_event_code%(-(len(events)-i))
                else:
                    self.code += self._event_code%(e.to_code(i), -(len(events)-i), i, a1, c1, d1)

        self.code += self._footer_code
            
        exec(compile(self.code, 'crtbp_ode_generated', mode='exec'))
        self.py_ode = locals()['crtbp_ode']
        self.right_part = \
        compiler.compile_isolated(self.py_ode,
                          [types.double, types.double[:], types.double[:]], 
                          return_type=types.double[:]).entry_point
                                  
    def build_model(self, 
                    const_set=None,
                    integrator=None,
                    solout=None):
        return crtbp3_custom_model(right_part=self.right_part,
                                   ev_size=len(self.events),
                                   integrator=integrator, 
                                   solout=solout, 
                                   stm=self.stm)                              



# -*- coding: utf-8 -*-
"""
Created on Sun Jan 27 23:07:24 2019

@author: stasb
"""

import numpy as np
import pandas as pd
from itertools import product
from functools import reduce
from .models import base_model
from .directions import base_direction, x_direction
from .events import event_detector

class manifold:
    '''
    Class manifold provides calculation of stable and unstable manifolds 
    for specified unstable orbit (central manifold) near Lagrange points.
    '''
    labels = list(product(['stable', 'unstable'], ['positive', 'negative']))
    shorts = [x[0][0]+x[1][0] for x in labels]
    mapper = dict([(x[1],x[0]) for x in enumerate(shorts)])

#    mapper = {'sp':0, # stable branch with positive disturbance
#              'sn':1, # stable branch with negative disturbance
#              'up':2, # unstable branch with positive disturbance
#              'un':3} # unstable branch with negative disturbance    
    
    columns = ['br','tr']
    
    def __init__(self, model, orbit, tol=1e-12):
        if not isinstance(model, base_model):
            raise TypeError('model should be an instance of base_model')
        self.model = model
        
        ssize = self.model.get_state_size()
#        if isinstance(orbit, pd.DataFrame):
#            sz = orbit.shape[1]
        if orbit.shape[1] < ssize:
            raise ValueError('orbit states should be at least %r size'%ssize)
        self.orbit = orbit
        
        self.tol = tol
        self.reset()
        
    def reset(self):
        self.trajs = None
        self.evout = None
        
    def calc_manifold(self, 
                      disturbance=1e-10,
                      direction=x_direction(),
                      maxt=np.pi,
                      events=[],
                      verbose=True):
        self.reset()
#        if verbose:
#            print('Manifold calculation:')
        for i in range(4):
            self.calc_branch(i,  disturbance, direction, maxt, events, verbose)
        return self.trajs
        
    def calc_branch(self,
                    branch_idx,
                    disturbance=1e-10, 
                    direction=x_direction(), 
                    maxt=np.pi,
                    events=[],
                    verbose=True,
                    last_state='mint',
                    ret_df=True):
        '''
        Calculates manifold branch. Branch defined by disturbance sign and
        direction of propagation: forward or backward.
        Disturbs each state of orbit with specified disturbance in specified
        direction and propagates disturbed state up to maxt time or up to
        events.
        '''
        
        if not isinstance(direction, base_direction):
            raise TypeError('direction should be an instance of base_direction')
        
        if isinstance(branch_idx, str):
            if branch_idx not in self.shorts:
                raise ValueError('branch_idx should be one of: %r'%self.shorts)
            branch_idx = self.mapper[branch_idx]

        disturbance = (-1)**branch_idx*np.abs(disturbance)

        # stable manifold => backward propagation
        # unstable manifold => forward propagation
        maxt = (-1)**(branch_idx//2+1)*np.abs(maxt)
        print(maxt)
        if verbose:
            print('Manifold branch:', ' '.join(self.labels[branch_idx]), 
                  ', trajectories:', self.orbit.shape[0])

        if isinstance(self.orbit, pd.DataFrame):
#            arr = self.orbit.values[:,1:]
#            arr = self.orbit['x':'vz'].values#[:,1:]
            arr = self.orbit.values[:,1:]
        else:
            arr = self.orbit[:,1:]

        detector = event_detector(self.model, events, self.tol)

        tr_lst = []
        ev_lst = []
        for i, s in enumerate(arr):
            s1 = s.copy()
            s1[3:6] += disturbance * direction(0., s1)
            df, ev = detector.prop(s1, 0.0, maxt, ret_df=False, last_state=last_state)
            
            zdf = np.zeros((df.shape[0], 1), dtype=float)            
            df = np.hstack((zdf+branch_idx,zdf+i,df))
            tr_lst.append(df)
            
            if ev.shape[0] > 0:
                zev = np.zeros((ev.shape[0], 1), dtype=float)
                ev = np.hstack((zev+branch_idx,zev+i,ev))           
                ev_lst.append(ev)
                
            print(i, end=' ')
        print()
#        print('Trajectories calculated:', len(tr_lst))
        print('States count:', reduce(lambda a,x: a+x.shape[0], tr_lst, 0))
        
        if self.trajs is None:
            self.trajs = np.vstack(tuple(tr_lst))
        else:
            self.trajs = np.vstack((self.trajs,*tr_lst))
        
        if len(ev_lst) > 0:
            if self.evout is None:
                self.evout = np.vstack(tuple(ev_lst))
            else:
                self.evout = np.vstack((self.evout,*ev_lst))
            
        if ret_df:
            self.trajs = self.to_df(self.trajs)
            if len(ev_lst) > 0:
                self.evout = self.to_df(self.evout,
                                        columns=manifold.columns+
                                                detector.columns+
                                                detector.model.columns)
        return self.trajs, self.evout
    
    def to_df(self, arr, columns=None):#, index_col=None):
#        if index_col is None:
#            index_col = manifold.columns[0]
        if columns is None:
            columns = manifold.columns+self.model.columns
        return self.model.to_df(arr, columns)#, index_col)
    
    def split_data(self, data):
        '''
        Splits data by columns: ['br', 'tr'] and ['t', 'x', 'y' ,...]
        '''
        if isinstance(data, pd.DataFrame):
#            d = data.reset_index()
            return data[manifold.columns], data[self.model.columns]
        n = len(manifold.columns)
        return data[:,:n], data[:,n:]
        




# -*- coding: utf-8 -*-
"""
Created on Tue Jan  1 14:11:11 2019

@author: stasb
"""

from scipy.interpolate import interp1d

class base_extrapolator:
    '''
    Class base_extrapolator is a common interface for all extrapolation
    classes.
    '''
    def __call__(self, arg, history):
        '''
        Calculate extrapolated value at arg using history of states.
        '''
        pass
    
class const_extrapolator(base_extrapolator):
    '''
    Class const_extrapolator returns previous state (last state from history).
    '''
    def __call__(self, arg, history):
        if not history:
            raise RuntimeError('History is empty')
        return history[-1]
    
class linear_extrapolator(base_extrapolator):
    '''
    Class const_extrapolator uses two last states to calculate linear
    extrapolation.
    '''
    def __init__(self, arg_index=0, value_index=1):
        '''
        Constructor
        
        Parameters
        ----------
        arg_index : int
            Index of argument within states stored in history.
            
        value_index : int
            Index of value within states stored in history. This value
            should be extrapolated.
        '''
        self.arg_index = arg_index
        self.value_index = value_index
        
    def __call__(self, arg, history):
        '''
        Calculate extrapolated value.
        
        Parameters
        ----------
        
        arg : float
            Argument where linear extrapolation should be calculated.
            
        history : reference to iterable
            States stored in history will be used for extrapolation.
        
        Returns
        -------
        value : float
            Extrapolated value. Linear extrapolation calculated
            using 2 last states stored in history. If there are
            less than 2 states in history then returns last state.
            
        See also
        --------
        scipy.interpolate.interp1d
        
        '''
        if not history:
            raise RuntimeError('History is empty')
        if len(history) < 2:
            return history[-1][self.value_index]
        intrp = interp1d([history[-2][self.arg_index], history[-1][self.arg_index]], 
                         [history[-2][self.value_index], history[-1][self.value_index]],
                         kind='linear',
                         fill_value='extrapolate')
        return intrp(arg)
    
class quadratic_extrapolator(linear_extrapolator):
    def __init__(self, arg_index=0, value_index=1, default='linear'):
        super().__init__(arg_index, value_index)
        self.default = (1 if default == 'linear' else 0)

    def __call__(self, arg, history):
        '''
        Calculate extrapolated value.
        
        Parameters
        ----------
        
        arg : float
            Argument where quadratic extrapolation should be calculated.
            
        history : reference to iterable
            States stored in history will be used for extrapolation.
        
        Returns
        -------
        value : float
            Extrapolated value. Quadratic extrapolation calculated
            using 3 last states stored in history. If there are
            less than 3 states in history then returns linear
            extrapolation (if default=='linear') or last state (else).
            
        See also
        --------
        scipy.interpolate.interp1d
        
        '''
        if not history:
            raise RuntimeError('History is empty')
        elif len(history) < 3:
            return super().__call__(arg, history)
        
        intrp = interp1d([history[-3][self.arg_index], history[-2][self.arg_index], history[-1][self.arg_index]], 
                         [history[-3][self.value_index], history[-2][self.value_index], history[-1][self.value_index]],
                         kind='quadratic',
                         fill_value='extrapolate')
        return intrp(arg)

class cubic_extrapolator(quadratic_extrapolator):
    def __init__(self, arg_index=0, value_index=1, default='quadratic'):
        super().__init__(arg_index, value_index)
        self.default = (1 if default == 'quadratic' else 0)

    def __call__(self, arg, history):
        if not history:
            raise RuntimeError('History is empty')
        elif len(history) < 4:
            return super().__call__(arg, history)
        
        intrp = interp1d([history[-4][self.arg_index], history[-3][self.arg_index], history[-2][self.arg_index], history[-1][self.arg_index]], 
                         [history[-4][self.value_index], history[-3][self.value_index], history[-2][self.value_index], history[-1][self.value_index]],
                         kind='cubic',
                         fill_value='extrapolate')
        return intrp(arg)



# -*- coding: utf-8 -*-
"""
Created on Fri Dec  7 12:22:48 2018

@author: stasb
"""

import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import brent, bisect

from .models import base_model
from .events import event_detector
from .directions import base_direction

class base_correction:
    '''
    Base interface class for all methods that calculate
    delta-V (correction impulse) for elimination of unstable component of
    motion in the vicinity of unstable Lagrange points.
    '''
    
    def __init__(self, model, direction):
        if not isinstance(model, base_model):
            raise TypeError('model should be an instance of base_model')
        self.model = model
        if not isinstance(direction, base_direction):
            raise TypeError('direction should be an instance of base_direction')
        self.direction = direction
        
    def calc_dv(self, t, s):
        pass

class border_correction(base_correction):
    '''
    Calculates delta-V using left and right borders.
    For example, borders are two planes orthogonal to x axis: one located 
    to the left of unstable point and one - to the right.
    See S.A. Aksenov, S.A. Bober "Calculation and Study of Limited Orbits
    around the L2 Libration Point of the Sun–Earth System". Cosmic Research, 
    2018, Vol. 56, No. 2, pp. 144–150.
    '''
    error_borders_unreachable = "Borders unreachable"
    error_before_unreachable = "Cant't reach event(s) before borders"
    error_max_iterations_reached = "Maximum iterations number reached"
    error_root_separation = "Can't separate discontinuity point"
    
    def __init__(self, model, direction,
                 left, right, before=[],
                 iterations=(20,100), maxt=10*np.pi,
                 dv0 = 1e-12,
                 tol=1e-16):
        '''
        Constructor
        '''
        super().__init__(model, direction)
        self.before = before
        if type(self.before) not in (list, tuple):
            self.before = (self.before,)
        self.left = left
        if type(self.left) not in (list, tuple):
            self.left = (self.left,)
        self.right = right
        if type(self.right) not in (list, tuple):
            self.right = (self.right,)
        self.iterations=iterations
        self.maxt = maxt
        self.dv0 = dv0
        self.tol = tol        

    def calc_dv(self, t, s):
        '''
        Calculate magnitude of delta-v correction vector directed
        along self.direction.
        
        Parameters
        ----------
        t : float
            Time when correction should be made
            
        s : numpy array
            State 
            
        Return
        ------
        dv : numpy array
            Delta-v vector embedded in model zero state.
            Correction maneuver can be made like this:
                s += dv
            
        '''
        if self.before:
            detector_before = event_detector(self.model, self.before, tol=self.tol)
        detector = event_detector(self.model, self.left+self.right, tol=self.tol)
        
        s1 = s.copy()
        vstart = s[3:6].copy()
        dv = self.dv0
        d = self.direction(t, s)

        def get_border(v):
            s1[3:6] = vstart+v*d
            if self.before:
                _, ev0 = detector_before.prop(s1, t, t+self.maxt, ret_df=False)
                if ev0.shape[0] < 1:
                    raise RuntimeError(self.error_before_unreachable)
                s2 = ev0[-1,4:]
            else:
                s2 = s1
            _, ev = detector.prop(s2, t, t+self.maxt, ret_df=False)
            if ev.shape[0] < 1:
                raise RuntimeError(self.error_borders_unreachable)
            return -1 if ev[-1,0] < len(self.left) else 1

        # root separation
        for i in range(self.iterations[0]):  
            pa = get_border(-dv)
            pb = get_border( dv)
            if pa == pb:
                dv *= 10
            else:
                break
#        print('separation: %e, it=%d'%(self.dv0,i))
        if i == self.iterations[0]:
            raise RuntimeError(self.error_root_separation)
        
        # root calculation
        res = bisect(get_border, -dv, dv, 
                     xtol=self.tol, maxiter=self.iterations[1],
                     full_output=True)
        v = res[0]
#        print('bisection: %e, it=%d'%(v, res[1].iterations))
        r = self.model.get_zero_state()
        r[:] = 0
        r[3:6] = v * d
        return r
    
class max_time_correction(base_correction):
    error_borders_unreachable = "Borders unreachable"

    def __init__(self, model,
                 direction,
                 border,
                 iterations=500, maxt=10*np.pi,
                 dv0 = 0.1, tol=1e-16):
        super().__init__(model, direction)
        self.border = border
        if type(self.border) not in (list, tuple):
            self.border = (self.border,)
        self.iterations=iterations
        self.maxt = maxt
        self.tol = tol
        self.dv0 = dv0
        self.mode = 0

    def calc_dv(self, t, s):
        detector = event_detector(self.model, self.border, tol=self.tol)
        d = self.direction(t, s)

        def time2border(v, s, beta_n):
#            print(v)
            s1 = s.copy()
            s1[3:6] += v * d
            arr, ev = detector.prop(s1, t, t+self.maxt, ret_df=False)
#            plt.plot(arr[:,1], arr[:,2])
            if ev.shape[0] < 1:
                raise RuntimeError(self.error_borders_unreachable)
            return -ev[-1,3]
        
        v = brent(time2border, args=(s, d),
                  brack=(-self.dv0, self.dv0),
                  tol=self.tol, maxiter=self.iterations)
        
        res = self.model.get_zero_state()
        res[:] = 0
        res[3:6] = v * d
        return res        
       
#class border_correction_deprecated(base_correction):
#    error_borders_unreachable = "Borders unreachable"
#    error_max_iterations_reached = "Maximum iterations number reached"
#    error_root_separation = "Can't separate discontinuity point"
#    
#    def __init__(self, model, direction, 
#                 left, right,
#                 iterations=(20,100),
#                 maxt=10*np.pi,
#                 dv0=1e-12, 
#                 tol=1e-16):
#        super().__init__(model, direction)
#        self.left = left
#        if type(self.left) not in (list, tuple):
#            self.left = (self.left,)
#        self.right = right
#        if type(self.right) not in (list, tuple):
#            self.right = (self.right,)
#        self.iterations=iterations
#        self.maxt = maxt
#        self.dv0 = dv0
#        self.tol = tol        
#
#    def calc_dv(self, t, s):
#        detector = event_detector(self.model, self.left+self.right, tol=self.tol)
#        
#        s1 = s.copy()
#        vstart = s[3:5].copy()
#        dv = self.dv0        
#        d = self.direction(t, s)
#
#        def get_border(v):
#            s1[3:6] = vstart+v*d
#            _, ev = detector.prop(s1, t, t+self.maxt, ret_df=False)
#            if ev.shape[0] < 1:
#                raise RuntimeError(self.error_borders_unreachable)
#            return -1 if ev[-1,0] < len(self.left) else 1
#
#        # root separation
#        for i in range(self.iterations[0]):  
#            pa = get_border(-dv)
#            pb = get_border( dv)
#            if pa == pb:
#                dv *= 10
#            else:
#                break
##        print('separation: %e, it=%d'%(self.dv0,i))
#        if i == self.iterations[0]:
#            raise RuntimeError(self.error_root_separation)
#        # root calculation
#        res = bisect(get_border, -dv, dv, 
#                     xtol=self.tol, maxiter=self.iterations[1],
#                     full_output=True)
#        v = res[0]
##        print('bisection: %e, it=%d'%(dv, res[1].iterations))
#        r = self.model.get_zero_state()
#        r[:] = 0
#        r[3:6] = v * d
#        return r

#class border_correction_old(base_correction):
#    '''
#    Calculates delta-V using left and right borders.
#    For example, borders are two planes orthogonal to x axis: one located 
#    to the left of unstable point and one - to the right.
#    See S.A. Aksenov, S.A. Bober "Calculation and Study of Limited Orbits
#    around the L2 Libration Point of the Sun–Earth System". Cosmic Research, 
#    2018, Vol. 56, No. 2, pp. 144–150.
#    '''
#    error_borders_unreachable = "Borders unreachable"
#    error_max_iterations_reached = "Maximum iterations number reached"
#    
#    def __init__(self, model, left, right,
#                 iterations=100, maxt=10*np.pi,
#                 angles=(90, 0), dv0 = 0.1, 
#                 tol=1e-16):
#        '''
#        Initializes border_correction instance.
#        
#        Parameters
#        ----------
#        model : base_model instance
#            Intended for constructing event_detector that detects border
#            crossing by spacecraft.
#            
#        left, right : list(base_event, ...) or tuple(base_event, ...)
#            Lists of events that represents left and right borders.
#            
#        iterations : int
#            Maximum number of iterations allowed for delta-V calculation.
#            
#        maxt : float
#            Maximum time allowed for spacecraft to reach borders.
#            
#        angles : tuple of two floats
#            angles[0] for first correction and angles[1] - for the rest.
#            
#        dv0 : float
#            Initial step for delta-V calculation.
#            
#        tol : float
#            Absolute tolerance for delta-V value.
#        '''
#        super().__init__(model)
#        self.left = left
#        if type(self.left) not in (list, tuple):
#            self.left = (self.left,)
#        self.right = right
#        if type(self.right) not in (list, tuple):
#            self.right = (self.right,)
#        self.iterations=iterations
#        self.maxt = maxt
#        self.angles = angles
#        self.dv0 = dv0
#        self.tol = tol
#
#    def calc_dv(self, t, s):
#        '''
#        Calculates delta-V.
#        
#        Parameters
#        ----------
#        t : float
#            Initial time for delta-V calculation.
#            
#        s : numpy.ndarray
#            State vector of spacecraft at time t.
#            
#        Return
#        ------
#        dv : numpy.ndarray
#            Calculated delta-V.
#            Size of array is as model.get_zero_state() returns.
#        '''
#        detector = event_detector(self.model, self.left+self.right, tol=self.tol)
#
#        s1 = s.copy()
#        vstart = s1[3:5].copy()
#        dv = self.dv0
#        
#        rads = math.radians(self.angles[self.mode])
#        beta_n = np.array([math.cos(rads), math.sin(rads)])
#           
#        _, ev = detector.prop(s1, t, t+self.maxt, ret_df=False)
#        if ev.shape[0] < 1:
#            raise RuntimeError(self.error_borders_unreachable)
#        p = 0 if ev[-1,0] < len(self.left) else 1
#        s1[3:5] = vstart + dv * beta_n
#
#        _, ev = detector.prop(s1, t, t+self.maxt, ret_df=False)
#        if ev.shape[0] < 1:
#            raise RuntimeError(self.error_borders_unreachable)
#        p1 = 0 if ev[-1,0] < len(self.left) else 1
#        
#        if p == p1 and p == 1:
#            dv = -dv
#           
#        v = dv        
#        i = 0
#        while math.fabs(dv) > self.tol and i < self.iterations:
#            s1[3:5] = vstart + v * beta_n
#            _, ev = detector.prop(s1, t, t+self.maxt, ret_df=False)
#            if ev.shape[0] < 1:
#                raise RuntimeError(self.error_borders_unreachable)
#            p1 = 0 if ev[-1,0] < len(self.left) else 1
#         
#            if p1 != p:
#                v -= dv
#                dv *= 0.5
#    
#            v += dv
#            i += 1
#        if i == self.iterations:
#            raise RuntimeError(self.error_max_iterations_reached)
#        self.mode = 1
#        res = self.model.get_zero_state()
#        res[:] = 0
#        res[3:5] = v * beta_n
#        return res




# -*- coding: utf-8 -*-
"""
Created on Sun Jan 20 18:46:45 2019

@author: stasb
"""

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="orbipy",
    version="0.2.2",
    author="Bober S.A.",
    author_email="stas.bober@gmail.com",
    description="Extensive orbital dynamics research tool for CRTBP and beyond",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://bitbucket.org/stas_bober/orbipy",
    packages=setuptools.find_packages(),
    package_data={
            'orbipy': ['data/*.csv', 'data/families/*.csv', 'data/families/*.npy'],
    },
    license='MIT',
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=['numpy>=1.15.2', 
                      'scipy>=1.1.0',
                      'pandas>=0.23.4',
                      'matplotlib>=3.0.0',
                      'numba>=0.39.0',
                      ],
    python_requires='>=3.6',
)



# -*- coding: utf-8 -*-
"""
Created on Wed Nov 28 00:02:39 2018

@author: stasb
"""
import numpy as np
import pandas as pd
from scipy.interpolate import interp1d
import matplotlib.pyplot as plt
from .models import crtbp3_model
from .scaler import scaler
from .mapper import mapper


class plottable:
    def get_df(self):
        pass
    
    def plot_df(self, df, ax, projection, **kwargs):
        pass
    
class point(plottable):
    def __init__(self, name, pos, coords=('x','y','z')):
        self.df = pd.DataFrame([pos], columns=coords)
        self.name = name
        
    def get_df(self):
        return self.df
    
    def plot_df(self, df, ax, projection, **kwargs):
        p=projection.split('-')
        if p[0] in df.columns and p[1] in df.columns:
            c0, c1 = df[p[0]], df[p[1]]
            ax.plot(c0, c1, **kwargs)
            ax.text(c0, c1, '  '+self.name)
    
    def __repr__(self):
        return 'plottable point [%s]: %r'%(self.name, 
                list(zip(self.df.columns, self.df.values[0]))) 
    

class plotter:

    label_mapper = {'T': {
                        'm':'min',
                        'h':'hours',
                        'd':'days',
                        'w':'weeks',
                        'M':'months',
                        'Y':'years'
                        },
                    'L': {
                        'Mm': r'${\cdot}10^3$ km',
                        'Gm': r'${\cdot}10^6$ km',
                        'au': 'AU'}
                    }
                                       
    @staticmethod
    def get_label(units, category):
        if category is None:
            return 'unidentified'
        u = units[category][1]
        if category == 'V':
            l, t = u.split('/')
            return plotter.label_mapper['L'].get(l, l) + '/' + \
                   plotter.label_mapper['T'].get(t, t)
        return plotter.label_mapper[category].get(u, u)
    
    def __init__(self, 
                 time_units='d',
                 length_units='km',
                 velocity_units='km/s',                 
                 nondim='nd'):
        self.scaler = scaler()

    @classmethod
    def from_model(cls, model,
                        time_units='d',
                        length_units='km',
                        velocity_units='km/s',
                        nondim='nd'):
        if not isinstance(model, crtbp3_model):
            raise TypeError('Only crtbp3_model currently supported')
        me = cls()
        me.scaler = scaler.from_model(model, 
                                      time_units=(nondim, time_units),
                                      length_units=(nondim, length_units),
                                      velocity_units=(nondim+'/'+nondim, velocity_units),
                                      nondim=nondim)
        me.M = point(model.M, np.array([-model.mu,0.,0.]))
        me.m = point(model.m, np.array([model.mu1,0.,0.]))
        for i, L in enumerate(model.L):
            name = 'L%d'%(i+1)
            setattr(me, name, point(name, L))
        return me
    
    @staticmethod
    def create_axes(ax=None,
                    proj_count=0,
                    fsize=(5.,5.),
                    vertical=False):
        if proj_count < 1:
            raise ValueError("Projections wasn't specified")
        
#        second = False
        if ax is None:
            if vertical:
                fig, ax = plt.subplots(proj_count,1,
                                       figsize=(fsize[0],fsize[1]*proj_count))
            else:
                fig, ax = plt.subplots(1,proj_count,
                                       figsize=(fsize[0]*proj_count,fsize[1]))
        else:
#            second = True
            fig = (ax[0] if isinstance(ax, np.ndarray) else ax).figure
        
        return fig, ax

        
    
    def translate_scale(self, 
                        data,
                        centers={'x':0., 'y':0., 'z':0.}):
        cdata = data.copy()
        for c in centers:
            if isinstance(data, pd.DataFrame):
                if c in cdata:
                    cdata[c] -= centers[c]
            else:
                i = mapper.col2idx(c)
                if i < cdata.shape[1]:
                    cdata[:,i] -= centers[c]
        
        cdata = self.scaler.convert_data(cdata)
        return cdata
    
    def plot(self, 
             data, 
             ax=None, 
             projection='x-y',
#             color='r',
#             label=None,
             **kwargs):
        if ax is None:
            raise RuntimeWarning("plotter.plot: axes wasn't specified")
            return
        
        p = tuple(projection.split('-'))
        
        if isinstance(data, pd.DataFrame):
            h = data[p[0]]
            v = data[p[1]]             
        else:
            i, j = mapper.col2idx(p[0]), mapper.col2idx(p[1])
#            if data.ndim == 3:
#                h, v = data[:,i,:], data[:,j,:]
#            else:
            h, v = data[:,i], data[:,j]
        
        ax.plot(h, v, **kwargs) # color, label=label, 
            
    def plot_proj(self, data=None,
                  transcale=True,
                  projections=('x-y','x-z','y-z'),
                  centers={'x':0., 'y':0., 'z':0.},
                  plottables=[],
                  vertical=False,
                  colors='b',
                  ax=None,
                  fsize=(5., 5.),
                  grid=False,
                  xlabel=True,
                  ylabel=True,
                  legend=False,
                  **kwargs):
                
        n = len(projections)
        fig, ax = self.create_axes(ax, n, fsize, vertical)
                       
        if transcale and data is not None:
            cdata = self.translate_scale(data, centers)
        else:
            cdata = data
        
        for ip, prj in enumerate(projections):
            if n > 1 and ip >= ax.shape[0]:
                break
            p = tuple(prj.split('-'))
                        
            xlbl = '%s, %s'%(p[0], self.get_label(self.scaler.units, mapper.col2cat(p[0])))
            ylbl = '%s, %s'%(p[1], self.get_label(self.scaler.units, mapper.col2cat(p[1])))
#            ylbl2 = '%s, %s'%(mapper.col2cat(p[1]), 
#                              self.get_label(self.scaler.units, mapper.col2cat(p[1])))

            cur_ax = ax[ip] if isinstance(ax, np.ndarray) else ax

            if 'label' not in kwargs:
                kwargs['label'] = ylbl
            if 'color' not in kwargs:
                kwargs['color'] = colors[ip%len(colors)]
            
            if cdata is not None:
                self.plot(cdata, cur_ax, prj, **kwargs)                        

            for pltbl in plottables:
                df = pltbl.get_df()
                if df is not None:
                    df = self.translate_scale(df, centers)
                    pltbl.plot_df(df, cur_ax, prj, **kwargs)
            
            if xlabel: cur_ax.set_xlabel(xlbl)
            if ylabel: cur_ax.set_ylabel(ylbl)
#            if ylabel: cur_ax.set_ylabel(ylbl2 if second else ylbl)
            cur_ax.grid(grid)
            if legend: cur_ax.legend()
            
        fig.tight_layout()
        return ax        
    
    def plot_manifold(self, data,
                      method='fast',
                      pts=100,
                      projections=('x-y','x-z','y-z'),
                      centers={'x':0., 'y':0., 'z':0.},
                      vertical=False,
                      colors='b',
                      ax=None,
                      fsize=(5., 5.),
                      grid=False,
                      xlabel=True,
                      ylabel=True,
                      legend=False,
                      **kwargs):
        n = len(projections)
        fig, ax = self.create_axes(ax, n, fsize, vertical)

        cdata = self.translate_scale(data, centers)
        
        method = method.lower()
        
        if method == 'fast':
            self.plot_proj(cdata.values[:,2:],
                           transcale=False,
                           projections=projections,
                           centers=centers,
                           vertical=vertical,
                           colors=colors,
                           ax=ax, fsize=fsize, grid=grid,
                           xlabel=xlabel, ylabel=ylabel,
                           legend=legend, **kwargs)
        elif method == 'slow':
            for br_idx in cdata.br.unique():
                br = cdata[cdata.br==br_idx]
                for tr in br.tr.unique():
                    self.plot_proj(br[br.tr==tr].values[:,2:],
                                   transcale=False,
                                   projections=projections, 
                                   centers=centers,
                                   vertical=vertical, colors=colors,
                                   ax=ax, fsize=fsize, grid=grid,
                                   xlabel=xlabel, ylabel=ylabel,
                                   legend=legend, **kwargs)
        elif method == 'intr':
            for br_idx in cdata.br.unique():
                br = cdata[cdata.br==br_idx]
                trajs = []
                for tr_idx in br.tr.unique():
                    arr = br[br.tr==tr_idx].values[:,2:]
                    l = np.zeros(arr.shape[0])
                    l[1:] = np.cumsum(np.sum((arr[1:,1:4]-arr[:-1,1:4])**2, axis=1)**0.5)
                    intrp = interp1d(l, arr[:,1:], axis=0, kind='cubic',
#                                     assume_sorted=False, 
                                     fill_value='extrapolate')
                    t = np.linspace(l[0], l[-1], pts)
                    s = intrp(t)
                    trajs.append(np.hstack((t.reshape(-1,1), s)))
                branch = np.stack(tuple(trajs), axis=2)
                
                self.plot_proj(branch,
                               transcale=False,
                               projections=projections, 
                               centers=centers,
                               vertical=vertical, colors=colors,
                               ax=ax, fsize=fsize, grid=grid,
                               xlabel=xlabel, ylabel=ylabel,
                               legend=legend, **kwargs)
        return ax




# -*- coding: utf-8 -*-
"""
Created on Fri Nov 23 23:53:02 2018

@author: stasb
"""

class base_solout:
    def __init__(self):
        self.reset()
        
    def reset(self):
        self.out = []
        
    def __call__(t, s):
        return 0
    
class default_solout(base_solout):

    def __call__(self, t, s):
        self.out.append([t, *s])
        return 0
    
class event_solout(base_solout):
    '''
    Class event_solout plays two roles:
        - solout: aggregates all states trough integration process
        - event separation: for each event calculates states right 
            before event occur.
    See also: events.py
    '''
    def __init__(self, events):
        self._events = events
        self.reset()

    @property
    def events(self):
        return self._events
    
    @events.setter
    def events(self, events):
        self._events = events

    def reset(self):
        self.out = []
        self.evout = []
        self.evvals = []
        self.counters = None

    def __call__(self, t, s):
        
        terminal = False           
        cur_evvals = [event(t, s) for event in self._events]
    
        if not self.out:
            if not self.counters:
                self.counters = [event.count for event in self._events]
            self.out.append([t, *s])
            self.evvals.append(cur_evvals)
            return 0
            
        self.out.append([t, *s])
        self.evvals.append(cur_evvals)
    
        for i, event in enumerate(self._events):            
            cur_val = cur_evvals[i]
            prev_val = self.evvals[-2][i]
    
            f1 = (prev_val < 0) and (cur_val > 0) and ((event.direction == 1) or (event.direction == 0))
            f2 = (prev_val > 0) and (cur_val < 0) and ((event.direction == -1) or (event.direction == 0))
            if (f1 or f2) and ((self.counters[i] == -1) or (self.counters[i] > 0)):
                if self.counters[i] > 0:
                    self.counters[i] -= 1
#                print(t)
                self.evout.append([i, # event index
                    (-1 if self.counters[i]==-1 else event.count-self.counters[i]), # event trigger counter
                    len(self.out)-2, # state before event
                    ])
                if event.terminal and ((self.counters[i] == -1) or (self.counters[i] == 0)):
                    terminal = True
                
        if terminal:
            return -1
        
        return 0